// -*- Mode:C++ -*-
//Header:
//File: gthread.h
//Author: NODA Itsuki
//Date: 1998/2/6
//EndHeader:

//Note:
// gthread.h : general platform independent thread utility
//  [class]
//	GaeaThread
//	GaeaMutex
//	GaeaCondVar
//	GaeaMutexCondVar
//	GaeaThreadKey
//
//  [functions]
//	Yield()
//EndNote:

#ifndef _gthread_h_
#define _gthread_h_
////////////////////////////////////////////////////////////////////////
#include <iostream.h>
#ifdef THREADED
extern "C" {
#include <pthread.h>
} ;
#endif

#include "gbtype.h"

//======================================================================
// IThread
//
class GaeaThread {
   public:
#ifdef THREADED
      pthread_t id ;
#else
      Int id ;
#endif

      inline void set_self_id() {
#ifdef THREADED
	 id = pthread_self() ;
#else
	 id = 0 ;
#endif
      } ;

#ifdef THREADED
      inline void create(pthread_attr_t* attr,
			 void * (*func)(void *),
			 void * addr) 
      {
	 pthread_create(&id,attr,func,addr) ;
      } ;
#else
      inline void create(void* dummy,
			 void * (*func)(void *),
			 void * addr) {} ;
#endif
      inline void create(void * (*func)(void *),
			 void * addr) {
	 create(NULLPTR,func,addr) ;
      } ;

      inline void join(void **value_ptr = NULLPTR) {
#ifdef THREADED
	 pthread_join(id,value_ptr) ;
#endif
      } ;

      inline void EXIT(void* value) {
#ifdef THREADED
	 pthread_exit(value) ;
#else
	 id = id ;
#endif
      } ;

      inline void cancel() {
#ifdef THREADED
	 pthread_cancel(id) ;
#else
	 id = id ;
#endif
      } ;

      inline void override_by_self_id() {
	 cancel() ;
	 set_self_id() ;
      } ;
} ;

#ifdef Solaris 
#ifdef THREADED
inline void Yield() { sched_yield() ;} ;
#else
inline void Yield() {} ;
#endif
#else
inline void Yield() {} ;
#endif

//======================================================================
// my Mutex
//

class GaeaMutex {
   public:
#ifdef THREADED
      pthread_mutex_t entity ;
#endif

      inline GaeaMutex() { init() ; } ;

      inline ~GaeaMutex() { destroy() ; } ;

      inline void init() {
#ifdef THREADED
	 pthread_mutex_init(&entity,NULLPTR) ;
#endif
      } ;

      inline void destroy() {
#ifdef THREADED
	 pthread_mutex_destroy(&entity) ;
#endif
      } ;

      inline void lock() {
#ifdef THREADED
	 pthread_mutex_lock(&entity) ;
#endif
      } ;

      inline void unlock() {
#ifdef THREADED
	 pthread_mutex_unlock(&entity) ;
#endif
      } ;
} ;

//======================================================================
// my Cond Var
//

class GaeaCondVar {
   public:
#ifdef THREADED
      pthread_cond_t entity ;
#else
      Int dummy ;
#endif

      inline GaeaCondVar() { init() ; } ;
      inline ~GaeaCondVar() { destroy() ; } ;

      inline void init() {
#ifdef THREADED
	 pthread_cond_init(&entity,NULLPTR) ;
#endif
      } ;

      inline void destroy() {
#ifdef THREADED
	 pthread_cond_destroy(&entity) ;
#endif
      } ;

      inline void signal() {
#ifdef THREADED
	 pthread_cond_signal(&entity) ;
#endif
      } ;

      inline void broadcast() {
#ifdef THREADED
	 pthread_cond_broadcast(&entity) ;
#endif
      } ;

      inline Int wait(GaeaMutex& mutex) {
#ifdef THREADED
	 return pthread_cond_wait(&entity,&(mutex.entity)) ;
#else
	 return 0 ;
#endif
	 
      } ;

      inline Bool timedwait(GaeaMutex& mutex,
			    UInt sec, UInt msec = 0, UInt usec = 0) 
      {
#ifdef THREADED
	 struct timeval ctime ;

	 gettimeofday(&ctime,NULLPTR) ;

	 struct timespec etime ;

	 UInt eusec = msec * 1000 + usec + ctime.tv_usec ;
	 etime.tv_nsec = (eusec % 1000000) * 1000 ;
	 etime.tv_sec = sec + ctime.tv_sec + (eusec / 1000000) ;

	 Int r = pthread_cond_timedwait(&entity,&(mutex.entity),&etime) ;

	 if(!r) return True ;
	 else return False ;
#else
	 return True ;
#endif
      } ;
};
	 

//======================================================================
// GaeaMutexCondVar
//

class GaeaMutexCondVar {
   public:
      GaeaMutex mutex ;
      GaeaCondVar cond ;

      inline GaeaMutexCondVar() : mutex(), cond() {} ;
      inline ~GaeaMutexCondVar() { destroy() ; } ;

      inline void init() { mutex.init() ; cond.init() ; } ;

      inline void destroy() {
	 mutex.destroy() ;
	 cond.destroy() ;
      } ;

      inline void lock() { mutex.lock() ; } ;

      inline void unlock() { mutex.unlock() ; } ;
      
      inline Int wait() { return cond.wait(mutex) ; } ;

      inline Bool timedwait(UInt sec, UInt msec = 0, UInt usec = 0) {
	 return cond.timedwait(mutex,sec,msec,usec) ;
      } ;

      inline Bool timedlockwait(UInt sec, UInt msec = 0, UInt usec = 0) {
	 mutex.lock() ;
	 Bool r = timedwait(sec,msec,usec) ;
	 mutex.unlock() ;
	 return r ;
      } ;

      inline void lockwait() {
	 mutex.lock() ;
	 cond.wait(mutex) ;
	 mutex.unlock() ;
      } ;

      inline void signal() { cond.signal() ; } ;
      
      inline void broadcast() { cond.broadcast() ; } ;

} ;

//======================================================================
// GaeaSharedMutexCondVar
//

class GaeaSharedMutexCondVar {
   public:
      GaeaMutex *mutex ;
      GaeaCondVar cond ;

      inline GaeaSharedMutexCondVar() : cond() {} ;
      inline ~GaeaSharedMutexCondVar() { destroy() ; } ;

      inline void init() { cond.init() ; } ;
      inline void init(GaeaMutex* mtx) { mutex = mtx ; cond.init() ; } ;

      inline void destroy() {
	 cond.destroy() ;
      } ;

      inline void lock() { mutex->lock() ; } ;

      inline void unlock() { mutex->unlock() ; } ;
      
      inline Int wait() { return cond.wait(*mutex) ; } ;

      inline Bool timedwait(UInt sec, UInt msec = 0, UInt usec = 0) {
	 return cond.timedwait(*mutex,sec,msec,usec) ;
      } ;

      inline Bool timedlockwait(UInt sec, UInt msec = 0, UInt usec = 0) {
	 mutex->lock() ;
	 Bool r = timedwait(sec,msec,usec) ;
	 mutex->unlock() ;
	 return r ;
      } ;

      inline void lockwait() {
	 mutex->lock() ;
	 cond.wait(*mutex) ;
	 mutex->unlock() ;
      } ;

      inline void signal() { cond.signal() ; } ;
      
      inline void broadcast() { cond.broadcast() ; } ;

} ;

//======================================================================
// GaeaThreadKey  (access method for thread local data)
//

class GaeaThreadKey {
   public:
#ifdef THREADED
      pthread_key_t entity ;
#else
      void * value ;
#endif

      inline Int init(void (*destructor)(void*) = NULLPTR) {
#ifdef THREADED
	 return pthread_key_create(&entity,destructor) ;
#else
	 value = NULLPTR ;
	 return 0 ;
#endif
      } ;

      inline Int destroy() {
#ifdef THREADED
	 return pthread_key_delete(entity) ;
#else
	 return 0 ;
#endif
      } ;

      inline void* get() {
#ifdef THREADED
	 return pthread_getspecific(entity) ;
#else
	 return value ;
#endif
      } ;

      inline Int set(const void *v) {
#ifdef THREADED
	 return pthread_setspecific(entity,v) ;
#else
	 value = (void *)v ;
	 return 0 ;
#endif
      } ;
} ;

////////////////////////////////////////////////////////////////////////
#endif /* _gthread_h_ */
