#ifndef _GSMessages_h_
#define _GSMessages_h_
/*
 *  File: GSMessages.h
 *  ------------------
 *  This defines the following classes (there is no .cc file).
 *      class GSHeardMessage
 *      class GSMessageRingBuffer
 *
 *  These store the messages a player hears.
 */

#include <libsclient.h>


//----------------------------------------------------------------------
// message

class GSHeardMessage {
public:
    ScCycleStep cycle ;
    ScSenderType sender ;
    ScString _buffer ;    // original 
    ScSubString message ; // current message
    ScFloat direction ;
    ScWord* pmode ;
    ScBool processp ;

    //--------------------------------------------------
    // creator
      
    GSHeardMessage() { init() ; } ;

    //--------------------------------------------------
    // init
      
    void init() {
        cycle.count = 0 ; cycle.update();
        _buffer.alloc(SC_BUFSIZE) ;
        direction = 0.0 ;
        pmode = SC_NULLPTR ;
        reset() ;
    } ;

    //--------------------------------------------------
    // reset
      
    void reset() {
        message.set(_buffer.head,_buffer.head) ;
        processp = ScFalse ;
    } ;

    //--------------------------------------------------
    // set

    void set(const ScHearInfo& hinfo) {
        reset() ;
        cycle.count = hinfo.time ; cycle.update() ;
        sender = hinfo.sender ;
        if(sender == SC_ST_referee) {
            pmode = hinfo.aux.pmode ;
        } else if (sender == SC_ST_player) {
            direction = hinfo.aux.direction ;
        }
        message.copy(ScSubString((char*)(hinfo._dummy_))) ;
    } ;
	 
    //--------------------------------------------------
    // processed

    void processed() { processp = ScTrue ; } ;

    //--------------------------------------------------
    // describe 

    void describe(ostream& ostr, const ScBool detailp = ScTrue,
                  const ScBool newlinep = ScTrue) const {
        ostr << "#GSHeardMessage[" << cycle.count << ":" 
             << scSenderName(sender) ;
        if(sender == SC_ST_player) 
            ostr << "(" << direction << ")" ; 
        ostr << ",msg=(" << message << ")/" << (int)processp << "]" ;
        if(detailp) {
            ostr << endl << "\ttime=" << cycle ;
            if(sender == SC_ST_referee)
                ostr << endl << "\tpmode=" << (void*)pmode ;
        }
        if(newlinep) ostr << endl ;
    } ;

    friend ostream& operator<< (ostream& ostr, const GSHeardMessage& dt) {
        dt.describe(ostr,ScFalse,ScFalse) ;
        return ostr ; 
    } ;
} ;


//----------------------------------------------------------------------
// message ring buffer

#define GS_MESSAGE_RING_BUFFER_SIZE	128

class GSMessageRingBuffer {
public:
    ScUInt head ;
    ScUInt tail ;
    ScUInt size ;
    GSHeardMessage* buffer ;
    ScMutex mutex ;

    //--------------------------------------------------
    // constructor

    GSMessageRingBuffer(ScUInt s = GS_MESSAGE_RING_BUFFER_SIZE) { 
        init(s) ; 
    } ;

    //--------------------------------------------------
    // init

    void init(ScUInt s) {
        head = 0 ;
        tail = 0 ;
        size = s ;
        buffer = new GSHeardMessage[size] ;
    } ;

    //--------------------------------------------------
    // nth

    GSHeardMessage* nth(const ScUInt idx) const {
        return &(buffer[idx % size]) ;
    } ;

    //--------------------------------------------------
    // get top/bottom

    GSHeardMessage* getTop() { return nth(head++) ; } ;

    GSHeardMessage* getBottom() { return nth(tail++) ; } ;
    GSHeardMessage* GetBottom() { 
        lock() ;
        GSHeardMessage* r = nth(tail++) ; 
        unlock() ;
        return r ;
    } ;

    //--------------------------------------------------
    // is emptry/full

    ScBool isFull() const { return head >= tail+size ; } ;
    ScBool isEmpty() const { return head <= tail ; };
    ScBool IsEmpty() { 
        lock() ; ScBool r = isEmpty() ; unlock() ;
        return r ; 
    } ;

    //--------------------------------------------------
    // is emptry/full
      
    void pushHearInfo(const ScHearInfo& hinfo) {
        lock() ;
        if(isFull()) {
            cerr << "Warning: Ring buffer is expired!" << endl ;
            getBottom() ;
        }
        GSHeardMessage* msg = getTop() ;
        msg->set(hinfo) ;
        unlock() ;
    } ;

    //--------------------------------------------------
    // lock/unlock

    void lock() { mutex.lock() ; }  ;
    void unlock() { mutex.unlock() ; } ;
      
    //--------------------------------------------------
    // describe

    void describe(ostream& ostr, const ScBool detailp = ScTrue,
                  const ScBool newlinep = ScTrue) const {
        ostr << "#GSMessageRingBuffer["  << size
             << ":" << head << "/" << tail 
             << "]" ;
        if(detailp) {
            for(ScUInt i = head ; i < tail ; i++) {
                ostr << endl << "\t" << i << ":" << *nth(i) ;
            }
        }
        if(newlinep) ostr << endl ;
    } ;

    friend ostream& operator<< (ostream& ostr, 
                                const GSMessageRingBuffer& dt) {
        dt.describe(ostr,ScFalse,ScFalse) ;
        return ostr ;
    } ;
} ;


// eof //
#endif
