/*
 *  File: GSMovingObject.cc
 *  -----------------------
 *  Along with GSMovingObject.h, defines:
 *      class GSMovingObject
 *      class GSMovingObjectTable
 *
 *  These are used to track moving objects.
 */
#include "GSMovingObject.h"


//----------------------------------------------------------------------
// Moving Object Database (entry)


ScBool GSMovingObject::scanSeenObj(const ScSeenObj& sinfo,
                                   const ScPosDir& origin,
                                   const ScPos& ovel,
                                   const ScInt t) {
    if(sinfo.know_dir()) {
        ScFloat angle = origin.hdir + sinfo.dir ;
        ScPos newPos(origin.pos.x + sinfo.dist * scCos(angle),
                     origin.pos.y + sinfo.dist * scSin(angle));
        
        // calculate velocity
        ScInt difference = t - ptime;
        if (difference > 0  &&
            difference <= maxTimeDiffForVelEstimate) {
            // calculate using position differences
            ScPos newVel = 
                EstimateZeroForceVelocity(isBall(), difference,
                                          posdir.pos,
                                          newPos, difference);
            if (difference <= maxTimeDiffForVelStamp) {
                setVel(newVel, t);
            } else {
                // iffy estimate, do not update timestamp.
                setVel(newVel);
            }
        } else {        
            if(sinfo.know_dir_chng()) {
                // if we have this data, calculate vel directly,
                // although it can be very noisy
                ScPos dv ;
                dv.x = sinfo.dist_chng * scCos(angle) 
                    - sinfo.dist * sinfo.dir_chng * scSin(angle) ;
                dv.y = sinfo.dist_chng * scSin(angle) 
                    + sinfo.dist * sinfo.dir_chng * scCos(angle) ;
                // iffy estimate, do not update timestamp.
                setVel(ovel.x + dv.x, ovel.y + dv.y) ;
            }
        }
        
        if(sinfo.know_head_dir()) {
            setDir(angle + sinfo.body_dir, angle + sinfo.head_dir, t) ;
        }

        // set the new position last
        setPos(newPos, t) ;
        return ScTrue ;
    } else {
        return ScFalse ;
    }
} 


ScBool GSMovingObject::ObstructeP(const ScFloat dir, 
                                  const ScFloat dist, const ScFloat width, 
                                  const ScPos org, 
                                  const ScInt time, 
                                  const ScInt tmargin) const {

    if(time - ptime > tmargin) return ScFalse ; 

    ScPos dpos = posdir.pos.relativePosFrom(org) ;
    ScPos rpos = dpos.Rotate(-dir) ;
    if(rpos.x < 0.0 || rpos.x > dist ||
       rpos.y < -width/2.0 || rpos.y > width/2.0) {
        return ScFalse ;
    } else {
        return ScTrue ;
    }
} 


//----------------------------------------------------------------------
// Moving Object Database (table)


ScBool GSMovingObjectTable::scanSeenObj(const ScSeenObj& sinfo,
                                        const ScPosDir& origin,
                                        const ScPos& ovel,
                                        const ScInt t) {
    if(sinfo.isBall()) {
        return ball.scanSeenObj(sinfo,origin,ovel,t) ;
    } else if(sinfo.isPlayer()) {
        if        (sinfo.inViewConeP && 
                   sinfo.id.player.isTeammate() &&
                   !sinfo.id.player.isUnknownUNum()) {
            GSMovingObject& p = getTeammate(sinfo.id.player.unum) ;
            return p.scanSeenObj(sinfo,origin,ovel,t) ;
        } else if (sinfo.inViewConeP && 
                   sinfo.id.player.isOpponent() &&
                   !sinfo.id.player.isUnknownUNum()) {
            GSMovingObject& p = getOpponent(sinfo.id.player.unum) ;
            return p.scanSeenObj(sinfo,origin,ovel,t) ;
        } else {
            GSMovingObject& p = getUnknown() ;
            p.id = sinfo.id.player ;
            return p.scanSeenObj(sinfo,origin,ovel,t) ;
        }
    } else {
        return ScFalse ;
    }
}


void GSMovingObjectTable::scanSeeInfo(const ScSeeInfo& sinfo,
                                      const ScPosDir& origin,
                                      const ScPos& ovel) {
    clearUnknown();
    for(ScInt i = 0 ; i < sinfo.n ; i++) {
        scanSeenObj(sinfo.obj[i],origin,ovel,sinfo.time) ;
    } 
} 


ScBool GSMovingObjectTable::isObstructed(const ScFloat dir, 
                                         const ScFloat dist, 
                                         const ScFloat width, 
                                         const ScPos org, 
                                         const ScInt time, 
                                         const ScInt tmargin) {
    for(ScUInt unum = 1 ; unum <= SC_N_PLAYER_PER_TEAM ; unum++) {
        const GSMovingObject& p = getTeammate(unum) ;
        if(p.ObstructeP(dir,dist,width,org,time,tmargin)) return ScTrue ;
    }
    for(ScUInt unum = 1 ; unum <= SC_N_PLAYER_PER_TEAM ; unum++) {
        const GSMovingObject& p = getOpponent(unum) ;
        if(p.ObstructeP(dir,dist,width,org,time,tmargin)) return ScTrue ;
    }
    for(ScUInt i = 0 ; i <= unknownN ; i++) {
        const GSMovingObject& p = getUnknown(i) ;
        if(p.ObstructeP(dir,dist,width,org,time,tmargin)) return ScTrue ;
    }
    return ScFalse ;
} 


// eof //
