#ifndef _GSMovingObject_h_
#define _GSMovingObject_h_
/*
 *  File: GSMovingObject.h
 *  ----------------------
 *  Along with GSMovingObject.cc, defines:
 *      class GSMovingObject
 *      class GSMovingObjectTable
 *
 *  These are used to track moving objects.
 */
#include <libsclient.h>
#include <assert.h>

ScPos EstimateZeroForceVelocity(const ScBool isBallP,
                                       const ScInt m,
                                       const ScPos& pos_t,
                                       const ScPos& pos_tn,
                                       const ScInt n) ;

static const ScInt GS_Default_Player_MaxTimeDiffForVelEstimate = 10 ;
static const ScInt GS_Default_Player_MaxTimeDiffForVelStamp = 5 ;
static const ScInt GS_Default_Ball_MaxTimeDiffForVelEstimate = 5 ;
static const ScInt GS_Default_Ball_MaxTimeDiffForVelStamp = 2 ;


//----------------------------------------------------------------------
// Moving Object Database (entry)

class GSMovingObject {
public:
    ScPlayerID id ;
    ScPosDir posdir ;
    ScInt ptime ;       // simTic when last calculated
    ScPos vel ;
    ScInt vtime ;       // simTic when last calculated

    ScInt maxTimeDiffForVelEstimate ;
    // Velocity is estimated using differences in position, this
    // parameter controls the maximum tick difference (interval) that
    // will be used.  Otherwise, velocity is estimated using component
    // change information.

    ScInt maxTimeDiffForVelStamp ;
    // The vtime is updated iff the tick difference is less than
    // this value.  
    
    //--------------------------------------------------
    // constructor

    GSMovingObject() {
        ptime = vtime = -1 ;
        maxTimeDiffForVelEstimate =
            GS_Default_Player_MaxTimeDiffForVelEstimate ;
        maxTimeDiffForVelStamp = 
            GS_Default_Player_MaxTimeDiffForVelStamp ;
    } ;

    //--------------------------------------------------
    // setPosDir

    ScPos& setPos(const ScFloat x, const ScFloat y, ScInt t = -1) {
        posdir.pos.x = x ; posdir.pos.y = y ;
        if(t >= 0) ptime = t ;
        return posdir.pos ;
    } ;
    ScPos& setPos(const ScPos src, ScInt t = -1) {
        if(t >= 0) ptime = t ;
        return posdir.pos.set(src) ; 
    } ;

    ScFloat& setBDir(const ScFloat bdir, ScInt t = -1) {
        if(t >= 0) ptime = t ;
        return posdir.bdir = bdir ;
    } ;
    ScFloat& setHDir(const ScFloat hdir, ScInt t = -1) {
        if(t >= 0) ptime = t ;
        return posdir.hdir = hdir ;
    } ;
    void setDir(const ScFloat bdir, const ScFloat hdir, ScInt t = -1) {
        setBDir(bdir) ; setHDir(hdir,t) ;
    } ;

    void setPosDir(const ScFloat x, const ScFloat y, 
                   const ScFloat bdir, ScInt t = -1) {
        setPos(x,y) ; setBDir(bdir,t) ;
    } ;
    void setPosDir(const ScPos pos, const ScFloat bdir, ScInt t = -1) {
        setPos(pos) ; setBDir(bdir,t) ;
    } ;

    ScPosDir& setPosDir(const ScFloat x, const ScFloat y, 
                        const ScFloat bdir, const ScFloat hdir, ScInt t = -1) {
        setPos(x,y) ; setDir(bdir,hdir,t) ;
        return posdir ;
    } ;
    ScPosDir& setPosDir(const ScPos pos, const ScFloat bdir, 
                        const ScFloat hdir, ScInt t = -1) {
        setPos(pos) ; setDir(bdir,hdir,t) ;
        return posdir ;
    } ;

    //--------------------------------------------------
    // setVel

    ScPos& setVel(const ScFloat vx, const ScFloat vy, ScInt t = -1) {
        vel.x = vx ; vel.y = vy ;
        if(t >= 0) vtime = t ;
        return vel ;
    } ;
    ScPos& setVel(const ScPos v, ScInt t = -1) {
        if(t >= 0) vtime = t ;
        return vel.set(v) ; 
    } ;

    //--------------------------------------------------
    // Ball or not

    const ScUInt BallIDTeam = SC_WT_other ;
      
    ScBool isBall() const { return id.team == SC_WT_other ; } ;

    void declareBall() {
        id.team = SC_WT_other ;
        maxTimeDiffForVelEstimate =
            GS_Default_Ball_MaxTimeDiffForVelEstimate ;
        maxTimeDiffForVelStamp = 
            GS_Default_Ball_MaxTimeDiffForVelStamp ;
    } ;

    //--------------------------------------------------
    // visual info to pos

    ScBool scanSeenObj(const ScSeenObj& sinfo,
                       const ScPosDir& origin,
                       const ScPos& ovel,
                       const ScInt t) ;

    //--------------------------------------------------
    // obstruction

    ScBool ObstructeP(const ScFloat dir, 
                      const ScFloat dist, const ScFloat width, 
                      const ScPos org, 
                      const ScInt time, const ScInt tmargin) const ;

    //--------------------------------------------------
    // distance to

    ScFloat distanceTo(const ScPos target) const {
        return scDistance(posdir.pos,target) ;
    } ;


    

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#GSMovingObject[" ;
        if(isBall()) { o << "ball" ;}
        else { o << id ; }
        o << ":" << posdir << "(" << ptime << "," << vtime << ")" << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const GSMovingObject& obj) {
        obj.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;
} ;

//----------------------------------------------------------------------
// Moving Object Database (table)

class GSMovingObjectTable {
public:
    GSMovingObject ball ;
    GSMovingObject teammate[SC_N_PLAYER_PER_TEAM] ;
    GSMovingObject opponent[SC_N_PLAYER_PER_TEAM] ;
    GSMovingObject unknown[SC_N_PLAYER] ;
    ScUInt unknownN ;

    //--------------------------------------------------
    // creator

    GSMovingObjectTable() { init() ; } ;

    //--------------------------------------------------
    // init

    void init() {
        ball.declareBall() ;
        for(ScUInt unum = 1 ; unum <=  SC_N_PLAYER_PER_TEAM ; unum++) {
            getTeammate(unum).id.team = SC_WT_our ;
            getTeammate(unum).id.unum = unum ;
        } ;
        for(ScUInt unum = 1 ; unum <=  SC_N_PLAYER_PER_TEAM ; unum++) {
            getOpponent(unum).id.team = SC_WT_opponent ;
            getOpponent(unum).id.unum = unum ;
        } ;
        for(ScUInt unum = 0 ; unum <  SC_N_PLAYER ; unum++) {
            unknown[unum].id.team = SC_WT_unknown ;
            unknown[unum].id.unum = SC_UNUM_unknown ;
        } ;
        unknownN = 0 ;
    } ;


    //--------------------------------------------------
    // access

    GSMovingObject& getBall() { return ball ; } ;
    GSMovingObject& getTeammate(const ScUInt unum) { 
        return teammate[unum-1] ; } ;
    GSMovingObject& getOpponent(const ScUInt unum) { 
        return opponent[unum-1] ; } ;
    GSMovingObject& getUnknown(const ScUInt idx) { 
        return unknown[idx] ; } ;
    GSMovingObject& getUnknown() { 
        unknownN++ ; 
        return unknown[unknownN-1] ; } ;
    void clearUnknown() { unknownN = 0 ; } ;

    //--------------------------------------------------
    // convert sensor info

    ScBool scanSeenObj(const ScSeenObj& sinfo,
                       const ScPosDir& origin,
                       const ScPos& ovel,
                       const ScInt t) ;

    void scanSeeInfo(const ScSeeInfo& sinfo,
                     const ScPosDir& origin,
                     const ScPos& ovel) ;

    //--------------------------------------------------
    // obstruction

    ScBool isObstructed(const ScFloat dir, 
                        const ScFloat dist, const ScFloat width, 
                        const ScPos org, 
                        const ScInt time, const ScInt tmargin) ;

    ScBool isObstructed(const ScPos dest,
                        const ScFloat dist, const ScFloat width, 
                        const ScPos org, 
                        const ScInt time, const ScInt tmargin) {
        return isObstructed(dest.relativePosFrom(org).Angle(),
                            dist,width,org,time,tmargin) ;
    } ;

    //--------------------------------------------------
    // nearest objects

    GSMovingObject* nearestObjectTo(const ScPos target,
                                    GSMovingObject *obj,
                                    const ScUInt n,
                                    const ScWhichTeam team,
                                    const ScUInt time,
                                    const ScUInt tmargin) {
        ScFloat min_dist = 2 * SC_PITCH_LENGTH ;
        ScInt min_i = -1 ;
        for(ScUInt i = 0 ; i < n ; i++) {
            const GSMovingObject& p = obj[i] ;
            if(p.id.team != team) continue ;
            if(time - p.ptime > tmargin) continue ;
            ScFloat dist = p.distanceTo(target) ;
            if(min_dist > dist) {
                min_i = i ;
                min_dist = dist ;
            }
        }
        if(min_i >= 0) return &(obj[min_i]) ;
        else return (GSMovingObject*)SC_NULLPTR;
    } ;


    GSMovingObject* nearestTeammateTo(const ScPos target,
                                      const ScUInt time, 
                                      const ScUInt tmargin) {
        GSMovingObject* p = nearestObjectTo(target, teammate, 
                                            SC_N_PLAYER_PER_TEAM,
                                            SC_WT_our, time, tmargin) ;
        GSMovingObject* q = nearestObjectTo(target, unknown, unknownN,
                                            SC_WT_our, time, tmargin) ;
        if      (scIsNullPtr(q)) 
            return p ;
        else if (scIsNullPtr(p))
            return q ;
        else if (p->distanceTo(target) < q->distanceTo(target)) 
            return p ;
        else
            return q ;
    } ;

    GSMovingObject* nearestOpponentTo(const ScPos target,
                                      const ScUInt time, 
                                      const ScUInt tmargin) {
        GSMovingObject* p = nearestObjectTo(target, opponent, 
                                            SC_N_PLAYER_PER_TEAM,
                                            SC_WT_opponent, time, tmargin) ;
        GSMovingObject* q = nearestObjectTo(target, unknown, unknownN,
                                            SC_WT_opponent, time, tmargin) ;
        if      (scIsNullPtr(q)) 
            return p ;
        else if (scIsNullPtr(p))
            return q ;
        else if (p->distanceTo(target) < q->distanceTo(target)) 
            return p ;
        else
            return q ;
    } ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#GSMovingObjectTable[" << this << "]" ;
        if(detailp) {
            o << endl ;
            o << "\t" << ball << endl ;
            for(ScUInt unum = 0 ; unum <  SC_N_PLAYER_PER_TEAM ; unum++) {
                o << "\t" << teammate[unum] << endl ;
            } ;
            for(ScUInt unum = 0 ; unum <  SC_N_PLAYER_PER_TEAM ; unum++) {
                o << "\t" << opponent[unum] << endl ;
            } ;
            for(ScUInt unum = 0 ; unum <  unknownN ; unum++) {
                o << "\t" << unknown[unum] << endl ;
            } ;
        } else {
            if(newlinep) o << endl ;
        }
    } ;

    friend ostream& operator<< (ostream& o, 
                                const GSMovingObjectTable& tbl) {
        tbl.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;
    
} ;	 


//--------------------------------------------------
// estimate velocity
// returns the value of velocity at time t+m, given
// 0-force assumption, and positions at time t and time
// t+n.  Typically, m>=n.
//
// The equation is Vtm = (Ptn - Pt) * (decay^m / sum(i=0 to n-1, decay^i)
//                           
// Note that the former value is a vector, the latter is a number.
//

inline ScPos EstimateZeroForceVelocity(ScBool isBallP,
                                       ScInt m,
                                       ScPos& pos_t,
                                       ScPos& pos_tn,
                                       ScInt n) {
    assert(m >= 0  &&  n >= 0);
    
    ScFloat decay = (isBallP ? SC_BALL_DECAY : SC_PLAYER_DECAY);
    ScFloat decaySum = 0;
    
    for (int i = 0; i < n; i++) {
        decaySum += pow(decay, i);
    }
    return (pos_tn - pos_t) * (pow(decay,m) / decaySum);
}



// eof //
#endif
