//
//  File: GSPlayer.cc
//  -----------------
//  Along with GSPlayer.h, defines:
//      class GSPlayer
//
//
#include "GSPlayer.h"



const ScFloat GSPlayer::GSCloseEnoughAngle;
const ScUInt GSPlayer::GSBallConfTicksPos;
const ScUInt GSPlayer::GSBallConfTicksVel;
const ScUInt GSPlayer::GSPlayerConfTicks;
const ScFloat GSPlayer::GSAvoidDistance;
const ScFloat GSPlayer::GSCloseMargin;
const ScFloat GSPlayer::GSCircleKickFactor;
const ScFloat GSPlayer::GSCarefulCircleKickFactor;
const ScFloat GSPlayer::GSMaxCircleKickPower;
const ScFloat GSPlayer::GSCircleKickCloseEnoughAngle;
const ScFloat GSPlayer::GSOptControlDist;
const ScFloat GSPlayer::GSBallVelDontCare;
const ScFloat GSPlayer::GSKickable;
const ScFloat GSPlayer::GSDashKickable;
const ScFloat GSPlayer::GSSmartKickFaceBallAngle;
const ScPos GSPlayer::GS_OurGoalCenter(-SC_PITCH_LENGTH/2.0, 0);
const ScFloat GSPlayer::GS_P0;
const ScFloat GSPlayer::GS_P1;
const ScFloat GSPlayer::GS_P2;
//V.1.06
//const ScFloat GSPlayer::GS_P3  = sqrt(scSq(SC_PENALTY_AREA_LENGTH) +
//                                      scSq(SC_PENALTY_AREA_WIDTH/2.0));
const ScFloat GSPlayer::GS_P3  = sqrt(scSq(SC_PENALTY_AREA_LENGTH) +
                                      scSq(SC_PENALTY_AREA_WIDTH/2.0)) ;
const ScFloat GSPlayer::GS_P4; //1.05o


GSPlayer::~GSPlayer() { }  // empty virual destructor


//--------------------------------------------------
// basic command stuff
//
// There is a slightly different strategy here.  It is assumed that
// the database is locked going into these commands, and will be
// unlocked at a higher level (calling function).  Similarly, waiting
// to send the command is done at the higher level.

ScBool GSPlayer::waitNextFootCommandTic() {
    ScTimeVal now;
    ScFloat diffMs;

    now.update();
    
    diffMs = now.MSecFloat() - (senseBodySignalP ?
                                bcycle.rtime.MSecFloat() :
                                waitCycle.rtime.MSecFloat());

    PDEBUG(cout << "\nwaitNextFootCommandTic: diffMs: "<<diffMs<<endl);

    //diffMS = now.MSec() -
    //    (senseBodySignalP ? bcycle.rtime.MSec() : fcycle.rtime.MSec());

    
    // if we last synced on bodySense, use bcycle, otherwise use waitCycle;

    if (diffMs >= simulationTic.MSecFloat()) {
        PDEBUG(cout << "waitNextFootCommandTic(): we're late!"<<endl);
        senseBodySignalP = ScFalse;
        waitCycle.update();
        return ScFalse;
    } else {
        diffMs = simulationTic.MSecFloat() - diffMs;
        PDEBUG(cout << "waitNextFootCommandTic(): ms: "<<diffMs<<endl);
        // wait for signal from sense-body (recvScanSInfo())
        //if (senseBodyFlag.timedwait(0, diffMs)) {  // 1.05o comment out
        if (senseBodyFlag.timedlockwait(0, diffMs)) {  // 1.05o
            PDEBUG(cout << "waitNextFootCommandTic(): signaled" <<endl);
            senseBodySignalP = ScTrue;
            waitCycle.update();
            return ScTrue;
        } else {
            PDEBUG(cout << "waitNextFootCommandTic(): timed out" << endl);
            senseBodySignalP = ScFalse;
            waitCycle.update();
            return ScFalse;
        }
    }
}


ScBool GSPlayer::waitNextFootCommandTicNoMark() {
    ScTimeVal now;
    ScFloat diffMs;
    now.update();
    // if we last synced on bodySense, use bcycle, otherwise use waitCycle;
    diffMs = now.MSecFloat() - (senseBodySignalP ?
                                bcycle.rtime.MSecFloat() :
                                waitCycle.rtime.MSecFloat());
    if (diffMs >= simulationTic.MSecFloat()) {
        return ScFalse;
    } else {
        diffMs = simulationTic.MSecFloat() - diffMs;
        // wait for signal from sense-body (recvScanSInfo())
        //if (senseBodyFlag.timedwait(0, diffMs)) { // 1.05o comment out
        if (senseBodyFlag.timedlockwait(0, diffMs)) {  // 1.05o
            return ScTrue;
        } else {
            return ScFalse;
        }
    }
}


ScBool GSPlayer::turn(const ScFloat& direction) {
    commandWaitAndLock();
    ScBool result = noLockTurn(direction);
    commandUnlock();
    return result;
}
    

ScBool GSPlayer::noLockTurn(const ScFloat& direction) {
    //if(scAbs(direction) < 1.0) return ScTrue;
    ScBool r = sendTurn(direction);
    if(r) {
        //
        // update position/direction
        //
        fcycle.update();
        lastAction = GSTurn;
        xposdir.bdir =
            scNormalizeAngle(xposdir.bdir + actualTurnAngle(direction));
    }
    // update position, velocity, stamina
    recoverEstimateStamina();
    xposdir.pos += xvel;
    xvel *= SC_PLAYER_DECAY;
    xcycle.update();
    return r;
}


ScBool GSPlayer::dash(const ScFloat& pow) {
    commandWaitAndLock();
    ScBool result = noLockDash(pow);
    commandUnlock();
    return result;
}


ScBool GSPlayer::noLockDash(const ScFloat& pow) {
    ScBool r = sendDash(pow);
    if(r) {
        fcycle.update();
        lastAction = GSDash;
        updateDashValues(pow);  // xposdir.pos, xvel, stamina, recovery, effort
    } else {
        // dash didn't happen, so update position and velocity accordingly
        xposdir.pos += xvel;
        xvel *= SC_PLAYER_DECAY;
        recoverEstimateStamina();
    }
    // update/recover stamina
    xcycle.update();
    return r;
};


// 
void GSPlayer::updateDashValues(const ScFloat& pow)
{
    ScFloat& stamina = estimateBodyState.stamina;
    ScFloat& effort = estimateBodyState.effort;

    // new x-values
    ScFloat depByPow = dashEffectivePower(pow) * SC_DASH_POWER_RATE;
    ScPos movement = xvel + Polar2ScPos(depByPow, xposdir.bdir);
    // there should be a cap on this, but the default server
    // values ensure that the player can never go over
    // SC_PLAYER_SPEED_MAX, so we don't need to worry about it
    xposdir.pos += movement;
    xvel = movement * SC_PLAYER_DECAY;

    // new stamina value
    stamina -= staminaConsumedByDash(pow);

    // new recovery value
    if (stamina <= SC_RECOVER_DEC_THR * SC_STAMINA_MAX) {
        recovery = scMax(SC_RECOVER_MIN, recovery - SC_RECOVER_DEC);
    }

    // new effort estimate
    if (stamina <= SC_EFFORT_DEC_THR * SC_STAMINA_MAX ) {
        effort = scMax(SC_EFFORT_MIN, effort - SC_EFFORT_DEC);
    } else if (stamina >= SC_EFFORT_INC_THR * SC_STAMINA_MAX ) {
        effort = scMin(1.0, effort + SC_EFFORT_INC);
    }

    // recover our stamina for next tick.
    recoverEstimateStamina();
}


ScBool GSPlayer::kick(const ScFloat& pow, const ScFloat& dir) {
    commandWaitAndLock();
    ScBool result = noLockKick(pow, dir);
    commandUnlock();
    return result;
}


ScBool GSPlayer::noLockKick(const ScFloat& pow, const ScFloat& dir) {
    ScBool r = sendKick(pow,dir);
    if(r) {
        fcycle.update();
        lastAction = GSKick;
        updateBallEstiInfo(pow, dir);
    }
    // update position, velocity, stamina (player only)
    xposdir.pos += xvel;
    xvel *= SC_PLAYER_DECAY;
    recoverEstimateStamina();
    xcycle.update();
    return r;
};


void GSPlayer::updateBallEstiInfo(ScFloat power, ScFloat rdir) {
    if (!KickableP()) {
        PDEBUG(cout << "\nupdateBallEstiInfo: not kickable?!?!"<<endl);
        return;
    }
    ScFloat adir = scNormalizeAngle(rdir + xposdir.bdir);
    ScFloat kickRate = SC_KICK_POWER_RATE * KickEffect();
    ScPos relMovement = Polar2ScPos(power*kickRate, adir);
    if (ballVelValid()) relMovement += ballVelAbs();
    estiBallPos = ballPosAbs() + relMovement;
    estiBallVel = relMovement * SC_BALL_DECAY;
    estiBallCycle.update();
    estiBallCycle.count = scycle.count + 1;
    // +1 since this info is for next cycle, see ballPosAbs();
}

// 1.04o
ScBool GSPlayer::catchBall(const ScFloat& dir) {
    commandWaitAndLock();
    ScBool r = sendCatch(dir);
    if (r) {
        lastAction = GSCatch;
        goalieLastCatchTime = scycle.count;
    }
    recoverEstimateStamina();  // 1.05
    commandUnlock();
    return r;
}


// 1.05
ScBool GSPlayer::turnTo(const ScFloat& dir) {
    commandLock();
    ScBool r = turnToRecursive(dir, estimateBodyState.speed);
    commandUnlock();
    return r;
}


// 1.05
ScBool GSPlayer::turnToRecursive(ScFloat adir, ScFloat speed) { 
    ScFloat relTurnAngle = scNormalizeAngle(adir - xposdir.bdir);
    ScFloat moment = relTurnAngle * (1.0 + SC_INERTIA_MOMENT * speed);
    if (moment > SC_MAXMOMENT  ||  moment < SC_MINMOMENT) {
        // need to do multiple turns to face the correct dir
        if (moment > 0) {
            noLockTurn(SC_MAXMOMENT);  // xpos updated for recursion
        } else {
            noLockTurn(SC_MINMOMENT);  // xpos updated for recursion
        }
        waitNextFootCommandTic();
        return turnToRecursive(adir, speed * SC_PLAYER_DECAY);
    }
    return noLockTurn(moment);
}


//--------------------------------------------------
// update object table


ScBool GSPlayer::recvScanSInfo() {
    if(socket.recv(recvBuffer)) {
        sinfo.scan(recvBuffer,ScTrue);
        switch(sinfo.type) {
            case SC_SIT_see:{
                scycle.set(sinfo.info.see.time);
                PDEBUG(ScFloat diff = vcycle.rtime.MSecFloat());
                vcycle.set(sinfo.info.see.time);
                PDEBUG(
                    diff -= vcycle.rtime.MSecFloat();
                    cout
                    << "\nVision: Time (ms): " << -diff
                    << "\nVision: NumObj: " << sinfo.info.see.n << endl;
                    );
                // if(estimatePosDirVel1()) {
                if(estimatePosDirVel2()) {
                    eposMutex.lock();
                    ecycle.set(sinfo.info.see.time);
                    ScBool fullUpdateP = updateXValuesFromEValues();
                    objectTableMutex.lock();
                    // use eposdir and evel, since xposdir and xvel
                    // might include actions not yet reflected in vision.
                    objectTable.scanSeeInfo(sinfo.info.see,eposdir,evel);
                    objectTableMutex.unlock();
                    eposMutex.unlock();
                    if (fullUpdateP) senseFullVisualFlag.signal();
                }
		//V.1.05noda
                senseVisualFlag.broadcast();
                //V.1.05
                for(ScInt i = 0 ; i < sinfo.info.see.n ; i++) {
                    if(sinfo.info.see.obj[i].isBall()) {
                        ballSeenCycle = sinfo.info.see.time ;
                        ballSeenRelDist = sinfo.info.see.obj[i].dist ;
                        ballSeenRelAngToBDir = sinfo.info.see.obj[i].dir ;
                        break ;
                    }
                }
                break; 
            }
            case SC_SIT_hear: {
                scycle.set(sinfo.info.hear.time);
                acycle.set(sinfo.info.hear.time);
                if(sinfo.info.hear.sender == SC_ST_referee) {
                    playmode = sinfo.info.hear.aux.pmode;
                }
                messageRingBuffer.pushHearInfo(sinfo.info.hear);
                break;
            }
            case SC_SIT_senseBody: {
                estimateBodyStateMutex.lock();
                adjustBodyState(sinfo);  // sets actual and esti bodyStates
                if (getLastAction() != GSDash) {
                    // scale xvel by speed
                    xposMutex.lock();
                    xvel = Polar2ScPos(estimateBodyState.speed, xvel.Angle());
                    xposMutex.unlock();
                }
                estimateBodyStateMutex.unlock();

                // send signal to waitNextFootCommandTic() 
                //senseBodyFlag.signal();
                senseBodyFlag.broadcast(); 
                break;   
            }
            case SC_SIT_error: {
                cerr << "Warning: previous command is " 
                     << (char*) sendBuffer << endl;
                break;
            }
            default: {
                return ScFalse;
            }
        }
        return ScTrue;
    } else {
        // if(socket.recv_errno == ECONNREFUSED) 
        // exit(1);
        return ScFalse;
    }
}


ScBool GSPlayer::updateXValuesFromEValues() {
    ScBool fullUpdateP = ScFalse;
    
    // newVel is in direction of evel, but with length of xvel
    ScPos newVel = Polar2ScPos(xvel.norm(), evel.Angle());

    // will need to eventually handle the case where
    // turning heads happen, in which case,
    // we trust the GSTurn stuff, minus whatever is
    // not trustworthy via the lastAction

    xposMutex.lock();

    xcycle = vcycle;

    // update our x-values: pos, bdir/hdir, vel
    switch (getLastAction()) {
        case GSNone: case GSError:
        case GSKick: case GSCatch: {  // 1.04o
            // all values can be trusted
            xposdir = eposdir;
            xvel = newVel;

	    //V.1.05noda
	    bdirLastUpdateByVisual = vcycle ;

            //cout << "#### XVal update (none/kick): pos, dir, vel"<<endl;
            //cout << "     XVal update (none/kick): "<<
            //    xposdir.pos << ", " << xposdir.bdir << ", " << xvel<<endl;

            fullUpdateP = ScTrue;
            
            break;
        }
        case GSDash: {
            // can trust direction
            xposdir.bdir = eposdir.bdir;
            xposdir.hdir = eposdir.hdir;

	    //V.1.05noda
	    bdirLastUpdateByVisual = vcycle ;

            // can't really trust position, but perhaps try it anyways
            //xposdir.pos = eposdir.pos;

            //cout << "#### XVal update (dash): dir"<<endl;
            //cout << "     XVal update (dash): "<< xposdir.bdir << endl;
            
            break;
        }
        case GSTurn: {
            // can trust position and velocity
            xposdir.pos = eposdir.pos;
            xvel = newVel;

	    //V.1.05noda
	    // if last update of bdir is too old, then trust visual
	    if(vcycle.count - bdirLastUpdateByVisual.count > 5) {
	       xposdir.bdir = eposdir.bdir;
	       xposdir.hdir = eposdir.hdir;
	       bdirLastUpdateByVisual = vcycle ;
	    }

            //cout << "#### XVal update (turn): pos, vel"<<endl;
            //cout << "     XVal update (turn): "<<
            //    xposdir.pos << ", " << xvel<<endl;

            break;
        }
        default: {
            cerr << "Bad Mojo." <<endl;
            exit(-1);
        }
    }
    xposMutex.unlock();
    return fullUpdateP;
}


ScInt GSPlayer::recvScanSInfoAll() {
    ScBool original_block_mode = socket.checkBlockingMode();
    if(original_block_mode) socket.setBlockingMode(ScFalse);

    ScInt c = 0;

    while(ScTrue) { 
        if(!recvScanSInfo()) break; 
        c++;
    }

    if(original_block_mode) socket.setBlockingMode(ScTrue);

    return c;
}


//--------------------------------------------------
// middle command




void GSPlayer::calcDistanceKick(GSCommandStep *pwo_cs,
                                ScFloat absKickDir,
                                ScFloat kickDist,
                                ScFloat dFactor,
                                ScFloat maxPower)
{
    ScFloat power;
    ScFloat newKickDist = kickDist * dFactor;
    ScFloat newKickDir = absKickDir;

    maxPower = scMin(maxPower, SC_MAXPOWER);

    // get the expected kick rate, this implicitly calls ballPosValid(),
    // since it is 0 if the ball is not kickable within BallConfTicks time
    ScFloat kickRate = SC_KICK_POWER_RATE * KickEffect();

    if (kickRate <= 0.0) {
        PDEBUG(cout<<"\ndistanceKick: not kickable."<<endl);
        calcError(pwo_cs);
        return;
    }

    PDEBUG(
        cout 
        <<"\ndistanceKick: infos."
        <<"\ndistanceKick:     dir:      "<<absKickDir
        <<"\ndistanceKick:     dist:     "<<kickDist
        <<"\ndistanceKick:     factor:   "<<dFactor
        <<"\ndistanceKick:     maxPow:   "<<maxPower
        <<"\ndistanceKick:     kickRate: "<<kickRate
        <<"\ndistanceKick:     bPosAbs:  "<<ballPosAbs()
        <<"\ndistanceKick:     bPosRel:  "<<ballPosRel()
        <<"\ndistanceKick:     bVelAbs:  "<<ballVelAbs()
        <<"\ndistanceKick:     bVelRel:  "<<ballVelRel()
        <<endl;
        );
    
    // calculate powers, correct for ball velocity
    if ( ballVelValid()  &&
         ballVelAbs().norm() > GSBallVelDontCare) {

        PDEBUG(
            cout << "\ndistanceKick: Correcting for ball vel.";
            cout << "\ndistanceKick:     old dir/dist: " << newKickDir
            << "/" << newKickDist << endl;
            );

        // get new vector for direction/distance in tmp
        ScPos tmp = Polar2ScPos(newKickDist, newKickDir) - ballVelAbs();
        newKickDir = tmp.Angle();
        newKickDist = tmp.norm();

        PDEBUG(cout << "\ndistanceKick:     new dir/dist: "
               << newKickDir << "/" << newKickDist << endl);
    }

    power = newKickDist / kickRate;
    
    if (power > maxPower) power = maxPower;

    PDEBUG(cout << "\ndistanceKick: calc power: " << power << endl);

    ScPos relMovement;
    if (ballVelValid()) {
        relMovement = ballVelAbs() + Polar2ScPos(power*kickRate, newKickDir);
    } else {
        relMovement = Polar2ScPos(power*kickRate, newKickDir);
    }

    calcKickTo(pwo_cs, power, newKickDir);
    //pwo_cs->nextBallPos = ballPosAbs() + relMovement;
    //pwo_cs->nextBallVel = relMovement * SC_BALL_DECAY;

    return;
}


// based on CMUnited turn_kick code.
void GSPlayer::calcCircleKick(GSCommandStep *pwo_cs, 
                              ScFloat absAngle, ScBool rightP,
                              ScBool stopBallP, ScFloat endDist,
                              ScFloat closeMargin, ScFloat kickFactor,
                              ScBool* pwo_DoneP,
                              ScBool controlCircleP)
{
    ScPos ktraj; // (ideal) trajectory of ball when kicked
    ScFloat kdir;
    ScFloat kdist;


    // if not kickable, or position is unknown, return false
    if (!KickableP()) {
        PDEBUG(cout << "circleKick: not kickable" << endl);
        if (pwo_DoneP != NULL) *pwo_DoneP = ScFalse;
        calcError(pwo_cs);
        return;
    }

    // if velocity is not known, turn to ball to get fix on it.
    if (!ballVelValid()) {
        PDEBUG(cout << "circleKick: unknown vel, turning." << endl);
        if (pwo_DoneP != NULL) *pwo_DoneP = ScFalse;
        calcTurnToAbs(pwo_cs, ballPosAbs());
        return;
    }
    
    PDEBUG(
        cout
        <<"\ncircleKick: infos."
        <<"\ncircleKick:     ballRelAngle: "<<ballAngleRel()
        <<"\ncircleKick:     desiredAngle: "<<absAngle
        <<"\ncircleKick:     diff: "<<absAngleDiff(ballAngleRel(),absAngle)
        <<"\ncircleKick:     ballRelPos: "<<ballPosRel()
        <<endl;
        );

     // if ball is in right spot, say so and return
    if (absAngleDiff(ballAngleRel(),absAngle) < GSCircleKickCloseEnoughAngle) {
        
        PDEBUG(
            cout<<"\ncircleKick: done.  Ball at: "<<ballAngleRel()<<endl;
            cout<<"circleKick: desired at: "<<absAngle<<endl;
            );

        if (pwo_DoneP != NULL) *pwo_DoneP = ScTrue; // mark done
        if (!stopBallP) {
            calcNoAction(pwo_cs);
            return;
        } else {
            PDEBUG(cout << "circleKick: Stopping ball." << endl);
            calcStopBallKick(pwo_cs);
            return;
        }
    }

    absAngle = scNormalizeAngle(absAngle);

    // 3 cases: (1) can kick directly to spot.
    //          (2) too close, kick at 90 deg in given direction.
    //          (3) not too close, kick along tangent in given direction.
    // get values for kdir and kdist.

    ScBool case2 = ScFalse; // yeah, bad style, but we need it


    if (straightKickP(absAngle, endDist, closeMargin)) {
        PDEBUG(cout << "\ncircleKick: case(1) kick directly to." << endl);
        // kick to the spot
        ktraj = Polar2ScPos(endDist, absAngle) - ballPosRel();
        kdir = scNormalizeAngle(ktraj.Angle());
        kdist = ktraj.norm();

        // don't kick over GSMaxCircleKickPower, these next 4 lines are new
        ScFloat kickRate = SC_KICK_POWER_RATE * KickEffect();
        ScFloat pow = kdist / kickRate;
        pow = scMin(pow, GSMaxCircleKickPower);
        kdist = pow * kickRate;
        
    } else if (distanceToBall() < closeMargin) {
        PDEBUG(cout << "\ncircleKick: case(2) kick 90." << endl);
        // kick at 90 deg, put it at the endDist (opt control dist)
        kdir = scNormalizeAngle(ballAngleRel() + (rightP ? 90 : -90));
        kdist = sqrt(scSq(endDist) - scSq(distanceToBall()));

        // note, we really want to get the ball out to endDist,
        // despite what the kickFactor may be, so divide by kickFactor,
        // but we need to wait until after the player movement correction
        case2 = ScTrue;

    } else {
        PDEBUG(cout << "\ncircleKick: case(3) kick tangent." << endl);
        // kick tangent, put it at endDist (opt control dist)
        kdir = 180.0 + ballAngleRel() +
            ((rightP ? -1 : 1) * scASin(closeMargin/distanceToBall()));
        kdir = scNormalizeAngle(kdir);
        kdist = sqrt(scSq(distanceToBall()) - scSq(closeMargin))
            + sqrt(scSq(endDist) - scSq(closeMargin));
        // this tends to require powerful kicks
    }

    PDEBUG(
        cout << "\ncircleKick misc: playerVel: "<<xvel<<endl;
        cout << "circleKick: pre-correction: dir: "<<kdir
             << " dist: "<<kdist<<endl;
        );

    playerMovementCorrectDirDist(&kdir, &kdist);

    PDEBUG(
        cout << "\ncircleKick: post-correction: dir: "<<kdir
             << " dist: "<<kdist<<endl;
        );

    if (case2) kdist /= kickFactor;  // see case2 kick above
    
    if (controlCircleP) {
        calcDistanceKick(pwo_cs,
                         kdir, kdist, kickFactor, GSMaxCircleKickPower);
    } else {
        // don't cap at GSMaxCircleKickPower
        calcDistanceKick(pwo_cs, kdir, kdist, kickFactor);
    }
}


// based from CMUnited code

// tests to see if the kick can be made without intersecting the
// closeMargin.  The degAngle and dist give the values where the ball
// is to be _after_ the kick, relative to the player (angle is not
// relative to player's direction)
ScBool GSPlayer::straightKickP(ScFloat degAngle, ScFloat dist,
                               ScFloat closeMargin)
{
    ScPos ballTraj = Polar2ScPos(dist, degAngle) - ballPosRel();
    // apply law of cosines:
    // c^2 = a^2 + b^2 - 2ab*cosC
    // let C be the angle, person-currentballPos-finalballPos
    //     a is dist, player-to-ball
    //     b is lenght of the ball trajectory
    //     c is dist
    // solving for C

    ScFloat a = distanceToBall();
    ScFloat b = ballTraj.norm();
    ScFloat c = dist;
    ScFloat C = scACos((scSq(c) - scSq(a) - scSq(b)) / (-2.0*a*b));
    
    // if obtuse, no way it can hit player
    if (C > 90.0) return ScTrue;

    // see how close it gets
    return ((a*scSin(C)) > closeMargin);
}


// based from playerMovementCorrection() in CMUnited code
void GSPlayer::playerMovementCorrectDirDist(ScFloat* pdir, ScFloat* pdist)
{
    // the relative target position for the ball.
    ScPos vRelBallTarg = ballPosRel() + Polar2ScPos(*pdist, *pdir);
    
    // the relative target position, corrected to be for where I'll be
    // next.
    ScPos vNewRelTarg = myPredictedPosition() - xposdir.pos +
        vRelBallTarg;
    
    // new trajectory for ball
    ScPos vNewRelTraj = vNewRelTarg - ballPosRel();
    
    *pdir  = vNewRelTraj.Angle();
    *pdist = vNewRelTraj.norm();
}


//--------------------------------------------------
// step Goto


ScBool GSPlayer::stepGoto(const ScPos& dest, ScFloat power, ScBool avoidp) {
    ScFloat dir;
    //cout << dest << eposdir << xposdir << endl;
    if(avoidp) {
        dir = findClearWay(dest,10.0,3.0,1.0,5);
        //ScFloat origdir = dest.relativePosFrom(xposdir.pos).Angle();
        //if(dir != origdir)
	//cout << "Avoid!! dir=" << dir 
	//	<< ", origdir=" << origdir
	//	<< ", diff=" << dir-origdir << endl;
    } else {
        dir = dest.relativePosFrom(xposdir.pos).Angle();
    }
    
    //cout << dir << power << endl ;

    turnTo(dir);
    dash(power);
    return ScTrue;
}



void GSPlayer::calcStepGoto2(GSCommandStep *pwo_cs, const ScPos& dest,
                             ScFloat power, ScBool avoidp) {
    ScFloat dir1, dir2;
    dir1 = dir2 = dest.relativePosFrom(xposdir.pos).Angle();
    
    if(avoidp) dir2 = findClearWay(dest,10.0,3.0,1.0,5);

    if ( absAngleDiff(dir1,dir2) < GSCloseEnoughAngle  &&
         absAngleDiff(xposdir.bdir,dir2) < GSCloseEnoughAngle) {
        // we're looking in the right direction, do a dash.

        //cout<<"calcStepGoto2: dashing!"<<endl;
        
        pwo_cs->action = GSDash;
        pwo_cs->power = power;
    } else {
        // turn to the right direction;

        //cout<<"calcStepGoto2: turning to "<< dir2 <<endl;
        
        calcTurnTo(pwo_cs, dir2);
    }
}
    

void GSPlayer::calcSmartKickTo(GSCommandStep* pwo_cs,
                               ScFloat kickPow,
                               ScFloat adir,                               
                               ScBool controlCircleP,
                               ScBool faceBallP,
                               GSRotationType rotType,
                               ScBool *pwo_DoneP)
{
    adir = scNormalizeAngle(adir);
    
    if (!ballPosValid()) {
        PDEBUG(cout << "\ncalcSmartKickTo: ball pos not valid."<<endl);
        if (pwo_DoneP != NULL) *pwo_DoneP = ScFalse;
        calcError(pwo_cs);
        return;
    }

    if (!KickableP()) {
        PDEBUG(cout << "\ncalcSmartKickTo: ball not kickable."<<endl);
        if (pwo_DoneP != NULL) *pwo_DoneP = ScFalse;
        calcError(pwo_cs);
        return;
    }

    if (infrontKickableTo(adir)) {
        PDEBUG(cout << "\ncalcSmartKickTo: in hemisphere."<<endl);

        // if flag is set, try to stop ball
        //
        // on 2nd though, don't bother, just use capped circleKicks.
        //
        //if (controlCircleP  &&  ballVelAbs().norm() > GSBallVelDontCare) {
        //    PDEBUG(cout << "\ncalcSmartKickTo: slowing down."<<endl);
        //    calcStopBallKick(pwo_cs);
        //    return;
        //}

        if (  faceBallP  &&
              scAbs(ballAngleRelBodyDir()) > GSSmartKickFaceBallAngle  ) {
            // see if ball will still be kickable if we take time to turn
            ScPos nextBallPosRel = ballPredictedPosition()
                - myPredictedPosition();
            if (nextBallPosRel.norm() < GSKickable) {
                PDEBUG(cout<<"\ncalcSmartKickTo: facing ball."<<endl);
                calcTurnToRel(pwo_cs, nextBallPosRel);
                return;
            }
            PDEBUG(cout<<"\ncalcSmartKickTo: can't face ball."<<endl);
        }
            
        PDEBUG(cout << "\ncalcSmartKickTo: doing final kick!"<<endl);
        if (pwo_DoneP != NULL) *pwo_DoneP = ScTrue;
        //calcKickTo(pwo_cs, kickPow, adir);
        calcKickTo2(pwo_cs, kickPow, adir);
        return;
    }

    // so, ball is in wrong hemisphere, try a circleKicking it
    // to the correct hemisphere.

    if (pwo_DoneP != NULL) *pwo_DoneP = ScFalse;
    ScBool rightP = getCorrectTurnDir(rotType, adir);

    PDEBUG(
        cout
        << "\nPassTo: circle-kicking:" 
        << "\n       ballAngleRel: " << ballAngleRel() 
        << "\n       TargetAngle:  " << adir 
        << "\n       RightP:       " << (rightP ? "T" : "F") << endl;
        );

    // just do normal circle-kicks
    calcCircleKick(pwo_cs, adir, rightP, ScFalse, NULL, controlCircleP);

    //if (controlCircleP) {
    //   // kick slower for more control
    //    calcCircleKick(pwo_cs, adir, rightP, ScFalse,
    //                   GSOptControlDist, GSCloseMargin,
    //                   GSCarefulCircleKickFactor, NULL);
    //} else {
    //    calcCircleKick(pwo_cs, adir, rightP, ScFalse,
    //                   GSOptControlDist, GSCloseMargin,
    //                   GSCircleKickFactor, NULL);
    //}
}


// returns TRUE iff "Rightwards" (clockwise) is the correct direction.
ScBool GSPlayer::getCorrectTurnDir(GSRotationType rotType,
                                   ScFloat targetADir)
{
    ScFloat nearHalfway, farHalfway;
    ScBool rightP = rightPCalculateHalfway(ballAngleRel(), targetADir,
                                    &nearHalfway, &farHalfway);
    
    switch (rotType) {
        case GSRotRight: {
            rightP = ScTrue;
            break;
        }
        case GSRotLeft: {
            rightP = ScFalse;
            break;
        }
        case GSRotClosest: {
            // Do nothing, rightP already in correct direction.
            break;
        }
        case GSRotFurthest: {
            rightP = !rightP;
            break;
        }
        case GSRotAvoid: {
            // v1.04o
            GSMovingObject* opp
                = objectTable.nearestOpponentTo(xposdir.pos,
                                                scycle.count,
                                                closestCheckTimeMargin);
            if (scIsNullPtr(opp) ||
                opp->distanceTo(xposdir.pos) > GSAvoidDistance) {
                // rotate closest; do nothing
                break;
            }
            
            ScFloat oppAngle = xposdir.pos.AngleTo(opp->posdir.pos);
            ScFloat ballAngle = ballAngleRel();
            // normalize to target direction
            oppAngle = scNormalizeAngle(oppAngle - targetADir);
            ballAngle = scNormalizeAngle(ballAngle - targetADir);

            if (oppAngle <= 0 && ballAngle >= 0 ||
                oppAngle >= 0 && ballAngle <= 0) {
                // on different sides of player relative to
                // target dir, just rotate ball to closest; do nothing
            } else {
                rightP = (oppAngle < ballAngle);
            }
            break;
        }
        default: {
            cerr << "Bad mojo." << endl;
            exit(-1);
            break;
        }
    }
    return rightP;
}


void GSPlayer::calcKickTo2(GSCommandStep *pwo_cs, ScFloat kickPow, ScFloat adir) {
    if (!KickableP()) {
        PDEBUG(cout<<"\ncalcKickTo2: not kickable."<<endl);
        calcError(pwo_cs);
        return;
    }
        
    if ( ballVelValid()  &&  ballVelAbs().norm() > GSBallVelDontCare) {

        ScFloat kickRate = KickEffect() * SC_KICK_POWER_RATE;
        ScFloat dist = kickPow * kickRate;

        PDEBUG(
            cout
            <<"\ncalcKickTo2: Correcting for ball vel."
            <<"\ncalcKickTo2:     ballVel:  "<<ballVelAbs()
            <<"\ncalcKickTo2:     old dir:  "<<adir
            <<"\ncalcKickTo2:     old dist: "<<dist
            <<"\ncalcKickTo2:     old pow:  "<<kickPow
            <<endl;
            );

        ScPos oldKick = Polar2ScPos(dist,adir);
        ScFloat refAngle = oldKick.Angle();

        // work with ref to the desired kick direction
        ScPos ballVelRef = ballVelAbs().Rotate(-refAngle);

        ScPos newKickRef;
        // newKickRef has enough y on it to cancel ball's y velocity.
        newKickRef.y = -ballVelRef.y;
        // and is just long enough to give the desired power
        if (dist > scAbs(newKickRef.y)) {
            newKickRef.x = sqrt(scSq(dist) - scSq(newKickRef.y));
        } else {
            newKickRef.x = 0;
            newKickRef.y = (newKickRef.y > 0) ? dist : -dist;
        }

        ScPos newKick = newKickRef.Rotate(refAngle);


        PDEBUG(
            cout
            <<"\ncalcKickTo2:     refAngle:   "<<refAngle
            <<"\ncalcKickTo2:     ballVelRef: "<<ballVelRef
            <<"\ncalcKickTo2:     newKickRef: "<<newKickRef
            <<"\ncalcKickTo2:     newKick:    "<<newKick
            <<endl;
            );

        
        adir = newKick.Angle();
        dist = newKick.norm();      // this is actually unneeded, LHS==RHS
        kickPow = dist / kickRate;  // this is actually unneeded, LHS==RHS
            
        PDEBUG(
            cout
            <<"\ncalcKickTo2:     new dir:  "<<adir
            <<"\ncalcKickTo2:     new dist: "<<dist
            <<"\ncalcKickTo2:     new pow:  "<<kickPow
            <<endl;
            );

        kickPow = scMin(kickPow, SC_MAXPOWER);
    }
    
    calcKickTo(pwo_cs, kickPow, adir);
}

// v1.04o made into "2" version
ScBool GSPlayer::getPlanForIntercept2(ScFloat dashPow,
                                      ScInt maxSteps,
                                      ScBool extraSafetyP,
                                      ScFloat margin,
                                      ScFloat ballVelFactor,
                                      ScFloat*pwo_absTurnDir,
                                      ScInt *pwo_numDashes,
                                      ScPos *pwo_endPlayerPos,
                                      ScPos *pwo_endBallPos,
                                      ScPos *pwo_endBallVel)
{
    if (!ballPosValid()  ||  !ballVelValid()) {
        PDEBUG(cout<<"getPlanForIntercept: ball pos/vel not valid!"<<endl);
        return ScFalse;
    }

    if (kickableP()) {
        PDEBUG(cout<<"getPlanForIntercept: ball is already kickable!"<<endl);
        if (pwo_absTurnDir != NULL) *pwo_absTurnDir = 0;
        if (pwo_numDashes != NULL) *pwo_numDashes = 0;
        return ScTrue;
    }
    
    // cache values, etc...
    ScPosDir orig_xposdir = xposdir;
    ScPos orig_xvel = xvel;
    ScFloat orig_stamina = estimateBodyState.stamina;
    ScFloat orig_effort = estimateBodyState.effort;
    ScFloat orig_recovery = recovery;

    // for purposes of calculation, assume we're standing still:
    xvel.x = xvel.y = 0;

    // init for i=1 step
    ScPos ballPos_i = ballPosAbs();
    ScPos ballVel_i = ballVelAbs() * ballVelFactor;
    ScFloat distToballPos_i = (orig_xposdir.pos - ballPos_i).norm();
    ScFloat dashDist_i_minus_1 = 0.0;
    ScFloat dashDist_i_minus_2 = 0.0;

    PDEBUG(
        cout
        <<"\n0 bPos, bVel: " << ballPos_i << ", " << ballVel_i
        <<"\n0 bDist, dDist-1, dDist-2: "
        <<    distToballPos_i << ", " << dashDist_i_minus_1
        <<    ", " << dashDist_i_minus_2
        <<"\n0 ballVTime: " << ball().vtime << endl;
        );

    int i;
    for (i = 1; i <= maxSteps; i++) {
        ballPos_i += ballVel_i;
        ballVel_i *= SC_BALL_DECAY;

        distToballPos_i = (orig_xposdir.pos - ballPos_i).norm();

        PDEBUG(
            cout
            <<"\n" << i << " bPos, bVel: "
            <<    ballPos_i << ", " << ballVel_i
            <<"\n" << i << " bDist, dDist-1, dDist-2: "
            <<    distToballPos_i << ", " << dashDist_i_minus_1
            <<    ", " << dashDist_i_minus_2 << endl;
            );
       
        // if we found a plan, calculate angle as just where the ball is
        // at in step i.  Note: this could lead to over-shooting, but
        // we'll just ignore that here.
        
        // perhaps want to do obstruction check and continue searching
        // if obstructed?

        if (extraSafetyP  &&
            distToballPos_i < dashDist_i_minus_2 + margin) {

            PDEBUG(cout << "Found safe plan." << endl);

            ScFloat absTurnDir = ballPos_i.AngleFrom(orig_xposdir.pos);
            if (pwo_absTurnDir != NULL)
                *pwo_absTurnDir = absTurnDir;
            if (pwo_numDashes != NULL)
                *pwo_numDashes = scMax((ScInt)0, (ScInt)(i - 2));
            if (pwo_endPlayerPos != NULL) 
                *pwo_endPlayerPos = orig_xposdir.pos
                    + Polar2ScPos(dashDist_i_minus_2, absTurnDir);
            if (pwo_endBallPos != NULL)
                *pwo_endBallPos = ballPos_i;
            if (pwo_endBallVel != NULL)
                *pwo_endBallVel = ballVel_i;
            break;
        }
        else if (!extraSafetyP  &&
                 distToballPos_i < dashDist_i_minus_1 + margin) {

            PDEBUG(cout << "Found normal plan." << endl);
            
            ScFloat absTurnDir = ballPos_i.AngleFrom(orig_xposdir.pos);
            if (pwo_absTurnDir != NULL)
                *pwo_absTurnDir = absTurnDir;
            if (pwo_numDashes != NULL)
                *pwo_numDashes = i - 1;
            if (pwo_endPlayerPos != NULL) 
                *pwo_endPlayerPos = orig_xposdir.pos
                    + Polar2ScPos(dashDist_i_minus_1, absTurnDir);
            if (pwo_endBallPos != NULL)
                *pwo_endBallPos = ballPos_i;
            if (pwo_endBallVel != NULL)
                *pwo_endBallVel = ballVel_i;
            break;
        }

        updateDashValues(dashPow); // pretend to do a dash
        dashDist_i_minus_2 = dashDist_i_minus_1;
        dashDist_i_minus_1 = (orig_xposdir.pos - xposdir.pos).norm();
    }

    // restore values
    xposdir = orig_xposdir;
    xvel = orig_xvel;
    estimateBodyState.stamina = orig_stamina;
    estimateBodyState.effort = orig_effort;
    recovery = orig_recovery;
    
    return (i <= maxSteps  &&  posInBoundsP(ballPos_i));  // 1.04o
}

// v1.04o
ScBool GSPlayer::getPlanForGoalieCatchIntercept(ScFloat *pwo_absTurnDir,
                                                ScInt *pwo_numDashes,
                                                ScFloat *pwo_absCatchDir)
{
    ScInt maxSteps = goalieMaxStepsForIntercept;
    ScFloat dashPow = goalieInterceptDashPower;

    ScFloat absTurnDir;
    ScInt numDashes;
    ScPos endPlayerPos, endBallPos;
    ScFloat margin;
    
    // attempt extra extra safe version first:
    //     bump ball velocity by +10%
    //     get in front of ball trajectory
    //     catch along trajectory (what about noise/overshoot?)
    //margin = SC_CATCHABLE_AREA_W / 4.0;
    //if (getPlanForIntercept2(dashPow, maxSteps, ScTrue, margin, 1.1, &absTurnDir,
    //                        &numDashes, &endPlayerPos, &endBallPos, NULL)) {
    //    if (endBallPos.x > -SC_PITCH_LENGTH / 2.0) {
    //        if (pwo_absTurnDir != NULL) *pwo_absTurnDir = absTurnDir;
    //        if (pwo_numDashes != NULL) *pwo_numDashes = numDashes;
    //       if (pwo_absCatchDir != NULL)
    //            *pwo_absCatchDir
    //                = scNormalizeAngle(ballVelAbs().Angle() + 180);
    //        PDEBUG(cout<<"Goalie: extra extra safe plan found"<<endl);
    //        return ScTrue;
    //    }
    //}

    // attempt std intercept, with large margin (-ball size for noise)

    margin = SC_CATCHABLE_AREA_L - SC_BALL_SIZE;
    if (getPlanForIntercept2(dashPow, maxSteps, ScFalse, margin,
                             1, &absTurnDir, &numDashes, &endPlayerPos,
                             &endBallPos, NULL)) {
        if (endBallPos.x > -SC_PITCH_LENGTH / 2.0) {
            if (pwo_absTurnDir != NULL) *pwo_absTurnDir = absTurnDir;
            if (pwo_numDashes != NULL) *pwo_numDashes = numDashes;
            if (pwo_absCatchDir != NULL)
                *pwo_absCatchDir
                    = endBallPos.AngleFrom(endPlayerPos);
            GDEBUG(cout
                   <<"getPlanForGoalieCatchIntercept: std plan found"<<endl);
            return ScTrue;
        }
    }
    GDEBUG(cout<<"getPlanForGoalieCatchIntercept: no plan found"<<endl);
    return ScFalse;
}


// Goalie stuff

// 1.04o
// dbase is locked upon entry,
//     returns LOCKED if no action (returns ScFalse),
//             UNLOCKED if action taken (returns ScTrue).
ScBool GSPlayer::attemptGoalieCatchIntercept(ScBool makeCatchP) {
    ScFloat aTurnDir, aCatchDir;
    ScInt numDash;
    GDEBUG(cout<<"attemptGoalieCatchIntercept: entering."<<endl);
    if (!getPlanForGoalieCatchIntercept(&aTurnDir, &numDash, &aCatchDir)
        ||  catchBanP(numDash + 1)) {
        // no plan found, or if there was one, the ball won't be
        // catchable at the time we get to it.
        GDEBUG(cout<<"attemptGoalieCatchIntercept: no plan found."<<endl);
        return ScFalse;
    }
    // unlock dbase, and execute plan
    GDEBUG(cout
           <<"attemptGoalieCatchIntercept: making intercept attempt."<<endl);
    
    commandUnlock(); 
    turnTo(aTurnDir);
    for (int i = 0; i < numDash; i++) {
        dash(goalieInterceptDashPower);
    }

    if (makeCatchP) {
        GDEBUG(cout
               <<"attemptGoalieCatchIntercept: making catch attempt."<<endl);
        ScFloat relBDir = scNormalizeAngle(aCatchDir - aTurnDir);
        catchBall(relBDir);
    }
    return ScTrue;
}


// 1.04o
// dbase is locked upon entry,
//     returns LOCKED if no action (returns ScFalse),
//             UNLOCKED if action taken (returns ScTrue).
ScBool GSPlayer::attemptGoalieKickIntercept(ScBool makeKickP) {
    ScFloat aTurnDir; ScInt numDash;
    ScFloat maxSteps = goalieMaxStepsForIntercept;
    ScFloat dashPow = goalieInterceptDashPower;
    GDEBUG(cout<<"attemptGoalieKickIntercept: entering."<<endl);
    if (!getPlanForIntercept(dashPow, maxSteps, ScFalse,
                             &aTurnDir, &numDash)) {  
        // no plan found
        GDEBUG(cout<<"attemptGoalieKickIntercept: no plan found."<<endl);
        return ScFalse;
    }
    // unlock dbase, and execute plan
    GDEBUG(cout
           <<"attemptGoalieKickIntercept: making intercept attempt."<<endl);
    commandUnlock(); 
    turnTo(aTurnDir);
    for (int i = 0; i < numDash; i++) {
        dash(dashPow);
    }

    if (makeKickP) {
        GDEBUG(cout<<"attemptGoalieKickIntercept: making kick attempt."<<endl);
        // just kick to closest outside direction
        ScFloat angle = 60;
        // should really lock dbase, but it's just a quick peek
        if (xposdir.pos.x < 0) angle *= -1;
        kickTo2(100, angle);
    }
    return ScTrue;
}


// 1.04o
// dbase is locked upon entry,
//     returns dbase UNLOCKED.
void GSPlayer::goalieToInterceptPos(ScInt perimeter) {
    static ScBool turnedBackToGoalForBackpedalP = ScFalse;
    ScPos bpos = ballPosAbs();
    
    ScFloat p;
    switch(perimeter) {
        case 0: p = GS_P0; break;
        case 1: p = GS_P1; break;
        case 2: p = GS_P2; break;
        case 3: p = GS_P3; break;
        default: cerr<<"Bad Mojo.\n"; exit(-1); break;
    }
    ScPos destPos = (Polar2ScPos(p, bpos.AngleFrom(GS_OurGoalCenter))
                     + GS_OurGoalCenter);
    ScFloat distToDestPos = scDistance(xposdir.pos, destPos);
    ScFloat distToGoal = scDistance(xposdir.pos, GS_OurGoalCenter);

    // check if already in position
    if (distToDestPos < 3.0) {
        //GDEBUG(cout<<"*"<<flush);
        //cout<<"Goalie: already in position."<<endl;
        turnedBackToGoalForBackpedalP = ScFalse;  // no longer backpedaling
        if (ballVelValid()) bpos = ballPredictedPosition();
        commandUnlock();
        turnToAbs(bpos);
        return;
    }

    // move into position, assuming goalie always has enough stamina/effort
    // (1) if destPos is in front 45 deg of view, turn (if needed)
    //     and dash once towards it.
    // (2) if we are inside P0, turn towards dest pos.
    // (3) otherwise: turn _once_ s.t. back is to CP,
    //     mark w/ static boolean
    //     on subsequent calls, backpedal at given speed
    //
    ScFloat dashPow;

    // (1)
    ScFloat angleDiffBDirToDestPos
        = absAngleDiff(xposdir.pos.AngleTo(destPos),
                       xposdir.bdir);
    if (angleDiffBDirToDestPos < 45) {
       //V.1.06 comment out
       //GDEBUG(cout<<"D"<<goalieMiscDashPower<<" "<<flush); //1.05o debug
        // destPos is "in front" of us
        turnedBackToGoalForBackpedalP = ScFalse; // no longer backpedaling
        dashPow = (distToDestPos
                   / (SC_DASH_POWER_RATE * estimateBodyState.effort));
        dashPow = scMin(goalieMiscDashPower, dashPow);
        commandUnlock();
        // turn, if needed
        if (angleDiffBDirToDestPos > 5) turnToAbs(destPos);
        dash(dashPow);
        return;
    }

    // (2)
    //if (distToGoal < GS_P0) { // 1.05o comment out
    if (distToGoal < GS_P0  ||  outOfFieldP) {
        GDEBUG(cout<<"T"<<flush);
        commandUnlock();
        turnedBackToGoalForBackpedalP = ScFalse; // no longer backpedaling
        turnToAbs(destPos);
        return;
    }

    // (3)
    ScFloat angleDiffBackToGoal
        = absAngleDiff(xposdir.pos.AngleTo(GS_OurGoalCenter),
                       xposdir.bdir+180); // our back is 180 deg from bdir
    // see if we need to turn s.t. our back is to the goal:
    //     we may be way off, so turn, or
    ///    we need to do that first turn to start our series of backdashes
    if (angleDiffBackToGoal > 45  
        ||  (turnedBackToGoalForBackpedalP == ScFalse  &&
             angleDiffBackToGoal > 5)) {
       // V.1.06: comment out
       //cout<<" T' "<<flush;
        turnedBackToGoalForBackpedalP = ScTrue;
        commandUnlock();
        turnTo(scNormalizeAngle(xposdir.pos.AngleTo(GS_OurGoalCenter)+180));
        return;
    }
    // backpedal
    GDEBUG(cout<<"B"<<flush);
    dashPow = (distToGoal
               / (SC_DASH_POWER_RATE * estimateBodyState.effort));
    dashPow = scMin(goalieMiscDashPower, dashPow);
    commandUnlock();
    dash(-dashPow);  // remember, we're facing away from our goal
    return;
        

#ifdef jfslfjlfdljdfl
    
    // back up to home position, if needed
    ScFloat homeDist = scDistance(GS_OurGoalCenter, xposdir.pos);
    ScUInt timeSinceLastBackup = scycle.count - lastBackupStep;
    if (homeDist > GS_P0  &&  timeSinceLastBackup > 3) {
        angle = scNormalizeAngle(180 + xposdir.pos.AngleTo(GS_OurGoalCenter));
        angleDiff = absAngleDiff(angle, xposdir.bdir);
        commandUnlock();
        if (angleDiff > 10) turnTo(angle);
        ScInt numDash = (ScInt)homeDist - 2;
        for (int i = 0; i < numDash; i++) {
            dash(-100); // can dash about a meter at 100 pow
        }
        lastBackupStep = scycle.count;
        return;
    }

    // we're already at home postion, do a dash outwards to
    // our intercept position
    angle = xposdir.pos.AngleTo(interceptPos);
    angleDiff = absAngleDiff(angle, xposdir.bdir);
    commandUnlock();
    if (angleDiff > 10) turnTo(angle);
    ScFloat dashPow
        = scMin(dist / (SC_DASH_POWER_RATE * estimateBodyState.effort),
                SC_MAXPOWER);
    dash(dashPow);
    return;

#endif
    
#ifdef jvklfldklds
    // basic action sequence: turn, dash, turn to ball
    GDEBUG(cout<<"Goalie: moving to intercept on P"<<perimeter<<endl);
    ScFloat dashPow
        = scMin(dist / (SC_DASH_POWER_RATE * estimateBodyState.effort),
                SC_MAXPOWER);
    ScFloat aTurnDir = interceptPos.AngleFrom(xposdir.pos);
    if (absAngleDiff(xposdir.bdir, aTurnDir) < 20) {
        // skip initial turn
        commandUnlock();
        dash(dashPow);
    } else if (absAngleDiff(xposdir.bdir, aTurnDir + 180) < 20) {
        // skip inital turn, dash backwards
        commandUnlock();
        dash(-dashPow);
    } else {
        // turn and dash
        commandUnlock();
        turnTo(aTurnDir);
        dash(dashPow);
    }

    // turn to ball, if necessary
    commandLock(); bpos = ballPosAbs(); commandUnlock();
    if (!(bpos.AngleFrom(xposdir.pos, xposdir.bdir) < 10)) {
        turnToAbs(bpos);
    }
    commandLock();
    return;
#endif
    
}

//V.1.05noda
ScCycleStep nodacheck ;
inline void NodaCheck(ScUInt id) {
   ScTimeVal ptime = nodacheck.rtime ;
   nodacheck.update() ;
   nodacheck.count++ ;
   cerr << "[" << id << ":" 
	<< ptime.diffMSec(nodacheck.rtime) << "ms]" << flush ;
} 
void nodaNop() { } ;
   

// 1.04o
// Goalie basic behavior when ball is in our side of the field.  Just
// a while(1) loop, with if()... statements.  LOCK dbase at top of
// loop, and UNLOCK prior to executing actions.
void GSPlayer::goalieCriticalActionLoop() {
    
    //GDEBUG(cout<<"Goalie: entering critical loop."<<endl);

   //V.1.05noda
   ScTimeVal ptime ;
   ptime.update() ;

    while (ScTrue) {

       //V.1.05noda
       {
	  ScTimeVal pptime ; pptime = ptime ;
	  ptime.update() ;
	  ScUInt msec = pptime.diffMSec(ptime) ;
	  if(msec < 100)  scSleep(0,100-msec) ;
       }
       
        // wait for the next tic, to give sensors change to catch up,
        // use no-mark version so we don't wait twice.
        waitNextFootCommandTicNoMark();  
        
        if (!goalieCriticalGameMode()) {
	   // V.1.06 : goalie always use narrow view
	   //sendChangeView(SC_VW_normal, SC_VQ_high);
            //GDEBUG(cout<<"Goalie: Non-critical mode, exiting loop."<<endl);
            return;
        }

        commandLock();
        ScPos bpos = ballPosAbs();
        
        //if (!ballPosValid()) {  //1.05o comment out
        if (!ballPosValid(5)) {
	   // V.1.06 : comment out
	   //GDEBUG(cout<<"Goalie(-): don't know where ball is!!"<<endl);
            commandUnlock();
	    //V.1.06 : goalie always use narrow view
            //sendChangeView(SC_VW_normal, SC_VQ_high);
            turnToAbs(bpos);  // guess it was there ...
            for (int i = 0; i < 6; i++) {
                waitForVisualInfo();
                if (ballPosValid()) {
                    cout<<"Goalie(-): found ball."<<endl;
		    //nodaNop() ;  // V.1.06 comment out
                    break;
                }
                turnTo(scNormalizeAngle(i*60 - 90));
            }
            GDEBUG(cout<<"Goalie(-): looked around for ball."<<endl);
            continue;
        }
        
        if (bpos.x > 0) {
            commandUnlock();
	    //V.1.06 : goalie always use narrow view
            //sendChangeView(SC_VW_normal, SC_VQ_high);
            //cout<<"Goalie: not on our side, exiting critical loop."<<endl;
            return;
        }

        GDEBUG(cout <<"$");

        // Case 1x: ball is close-in to our position
        if (goalieCriticalActionLoopBallIsClose()) {
            // action taken, dbase is unlocked by the above call
            continue;
        } else {
            ; // no action taken, dbase is still locked, fall through
        }

        // Case 2, ball inside P1, don't know velocity.  Run towards
        // it, hope that we get vel update soon.
        if (!ballVelValid()  &&  insideP1(bpos)) {
            GDEBUG(cout<<"Goalie(2): inside P1, don't "
                   "know vel!  StepDash.."<<endl);
            commandUnlock();
            // use intercept dash pow, since it's in P1
            stepDashToBall2(goalieInterceptDashPower, ScFalse);
            continue;
        }

        // Case 3, we know velocity, and ball is in P1 or in P2 and
        // vectored into goal.  This is pretty bad.  Just try blind
        // intercept/grab.  Nearcode should execute on next loop if
        // this doesn't work.
        //
        // Alternately, we may try to narrow vision, run to ball
        // traj+1, and hope that we can do an epos grab, but
        // this takes 3+ extra timesteps (1+ for beating ball, 2+ for
        // margin of 0).
        //
        if (ballVelValid()  &&
             (insideP1(bpos) ||
              (insideP2(bpos) && ballVelocityIntoMyGoalP() &&
               (ballVelAbs().norm() >= goalieFreeBallInP2MaxSpeed)))
            ) {
            GDEBUG(cout<<"Goalie(3): intercepting ball."<<endl);
            if (attemptGoalieCatchIntercept(ScTrue)) {
                // dbase unlocked
                continue;
            }
            // dbase still locked
            if (attemptGoalieKickIntercept(ScTrue)) {
                // dbase unlocked
                continue;
            }
            // dbase still locked
            GDEBUG(cout<<"Goalie(3): couldn't find intercept plan (sd)."<<endl);
            commandUnlock();
            stepDashToBall2(goalieMiscDashPower, ScFalse);
            continue;
            // no intercepts possible, not much to do, but fall
            // through anyways
        }
        
        // Case 4: ball is inside of P2, and not moving very fast.
        // in this case, we can dash right up to it, and let the
        // close-action code above handle things.  By bumping the
        // vision, we can be pretty sure that the epos stuff will
        // be used.
        //
        // alternate case 4, use attemptGoalieCatchIntercept(ScTrue),
        // and hope he just grabs it, which odds are pretty good,
        // since the ball isn't moving very fast.
        //
        /*
        if (  insideP2(bpos)
              && ballVelValid()
              && (ballVelAbs().norm() <
                  goalieFreeBallInP2MaxSpeed)  ) {
	    //V.1.06 : goalie always use narrow view
            //sendChangeView(SC_VW_narrow, SC_VQ_high);
            GDEBUG(cout<<"Goalie(4): dash right up "
                   "to a slow moving ball."<<endl);
            if (attemptGoalieKickIntercept(ScFalse)) {
                // dbase unlocked
                continue;
            } else {
                // dbase still locked
                GDEBUG(cout<<"Goalie(4): no plan on a "
                       "slow moving ball?!? (sd)"<<endl);
                commandUnlock();
                stepDashToBall2(100, ScFalse);
                continue;
            }
        }
        */

        // alternate case 4: as above, but lets just dash towards it,
        // since it isn't moving very fast.
        if (  insideP2(bpos)
              && ballVelValid()
              && (ballVelAbs().norm() <
                  goalieFreeBallInP2MaxSpeed)  ) {
            commandUnlock();
            // like an intercept
            GDEBUG(cout<<"Goalie(4): dash right up "
                   "to a slow moving ball."<<endl);
            stepDashToBall2(goalieInterceptDashPower, ScFalse);
            continue;
        }
        

        // Case 5, don't know vel, ball may be inside of P2, but we need
        // vel info to do calculations.
        if (!ballVelValid()) {
            GDEBUG(cout<<"Goalie(5): don't know ball velocity!!."<<endl);
            commandUnlock();
            turnToAbs(bpos);
	   //V.1.05noda
	    //NodaCheck(11) ;
	   senseVisualFlag.timedlockwait(1) ;
            continue;
        }

        // Case 6, ball inside P2, and not headed into goal.
        // Try fast/blind intercept only if odds look good.
        if (insideP2(bpos)  &&  ballFreenessInGoalArea() > .8) {
            GDEBUG(cout<<"Goalie(6): odds look good, intercept."<<endl);
            if (attemptGoalieCatchIntercept(ScTrue)) {
                // dbase unlocked
                continue;
            } else {
                GDEBUG(cout<<"Goalie(6): no plan, fall though."<<endl);
                ; // no plan found, dbase still locked, fall through
            }
        }

        if (!insideP3(bpos)) {
            //cout<<"Goalie(.): position on P1."<<endl;
            goalieToInterceptPos(1);  // unlocks dbase
	   //V.1.05noda
	    //NodaCheck(13) ;
	   senseVisualFlag.timedlockwait(1) ;
            continue;
        } else { 
            //cout<<"Goalie(.): position on P0."<<endl;
            goalieToInterceptPos(0);  // unlocks dbase
	   //V.1.05noda
	    //NodaCheck(14) ;
	   senseVisualFlag.timedlockwait(1) ;
            continue;
        }
        
    } // end of while(1)
}


// 1.05o
// 2nd version of this, no enclosing while(1) loop.
void GSPlayer::goalieCriticalActionLoop2() {
    
    //GDEBUG(cout<<"Goalie: entering critical loop."<<endl);
       
    // wait for the next tic, to give sensors change to catch up,
    // use no-mark version so we don't wait twice.
    waitNextFootCommandTicNoMark();  
        
    if (!goalieCriticalGameMode()) {
        //GDEBUG(cout<<"Goalie: Non-critical mode, exiting loop."<<endl);
       //V.1.06 : goalie always use narrow view
       // sendChangeView(SC_VW_normal, SC_VQ_high);
        waitNextFootCommandTic();  // do nothing
        return;
    }

    commandLock();
    ScPos bpos = ballPosAbs();
        
    if (!ballPosValid(5)) {
        GDEBUG(cout<<"Goalie(-): don't know where ball is!!"<<endl);
        commandUnlock();
	//V.1.06 : goalie always use narrow view
        //sendChangeView(SC_VW_normal, SC_VQ_high);
        turnToAbs(bpos);  // guess it was there ...
        for (int i = 0; i < 6; i++) {
            waitForVisualInfo();
            if (ballPosValid()) {
                cout<<"Goalie(-): found ball."<<endl;
                nodaNop() ;
                break;
            }
            turnTo(scNormalizeAngle(i*60 - 90));
        }
        GDEBUG(cout<<"Goalie(-): looked around for ball."<<endl);
        return;
    }
        
    if (bpos.x > 0) {
        //cout<<"Goalie: not on our side, exiting critical loop."<<endl;
        commandUnlock();
	//V.1.06 : goalie always use narrow view
        //sendChangeView(SC_VW_normal, SC_VQ_high);
        waitNextFootCommandTic();  // do nothing
        return;
    }

    GDEBUG(cout <<"$");

        // Case 1x: ball is close-in to our position
    if (goalieCriticalActionLoopBallIsClose()) {
        // action taken, dbase is unlocked by the above call
        return;
    } else {
        ; // no action taken, dbase is still locked, fall through
    }

    // Case 2, ball inside P1, don't know velocity.  Run towards
    // it, hope that we get vel update soon.
    if (!ballVelValid()  &&  insideP1(bpos)) {
        GDEBUG(cout<<"Goalie(2): inside P1, don't "
               "know vel!  StepDash.."<<endl);
        commandUnlock();
        // use intercept dash pow, since it's in P1
        stepDashToBall2(goalieInterceptDashPower, ScFalse);
        return;
    }

    // Case 3, we know velocity, and ball is in P1 or in P2 and
    // vectored into goal.  This is pretty bad.  Just try blind
    // intercept/grab.  Nearcode should execute on next loop if
    // this doesn't work.
    //
    // Alternately, we may try to narrow vision, run to ball
    // traj+1, and hope that we can do an epos grab, but
    // this takes 3+ extra timesteps (1+ for beating ball, 2+ for
    // margin of 0).
    //
    if (  ballVelValid()
          && (insideP1(bpos)
              || (insideP2(bpos) && ballVelocityIntoMyGoalP()
                  && (ballVelAbs().norm() >= goalieFreeBallInP2MaxSpeed)))
        ) {
        GDEBUG(cout<<"Goalie(3): intercepting ball."<<endl);
        if (attemptGoalieCatchIntercept(ScTrue)) {
            // dbase unlocked
            return;
        }
        // dbase still locked
        if (attemptGoalieKickIntercept(ScTrue)) {
            // dbase unlocked
            return;
        }
        // dbase still locked
        GDEBUG(cout<<"Goalie(3): couldn't find intercept plan (sd)."<<endl);
        commandUnlock();
        stepDashToBall2(goalieMiscDashPower, ScFalse);
        return;
        // no intercepts possible, not much to do
    }
        
    // Case 4: ball is inside of P2, and not moving very fast.
    // in this case, we can dash towards it.  When we get close
    // enough, we let the close-action code above handle things.
    if (  insideP2(bpos)
          && ballVelValid()
          && (ballVelAbs().norm() < goalieFreeBallInP2MaxSpeed)  ) {
        commandUnlock();
        // like an intercept
        GDEBUG(cout<<"Goalie(4): dash right up "
               "to a slow moving ball."<<endl);
        stepDashToBall2(goalieInterceptDashPower, ScFalse);
        return;
    }
        

    // Case 5, don't know vel, ball may be inside of P2, but we need
    // vel info to do calculations.
    if (!ballVelValid()) {
        GDEBUG(cout<<"Goalie(5): don't know ball velocity!!."<<endl);
        commandUnlock();
        turnToAbs(bpos);
        waitForVisualInfo();
        return;
    }

    // Case 6, ball inside P2, and not headed into goal.
    // Try fast/blind intercept only if odds look good.
    if (insideP2(bpos)  &&  ballFreenessInGoalArea() > .8) {
        GDEBUG(cout<<"Goalie(6): odds look good, intercept."<<endl);
        if (attemptGoalieCatchIntercept(ScTrue)) {
            // dbase unlocked
            return;
        } else {
            GDEBUG(cout<<"Goalie(6): no plan, fall though."<<endl);
            ; // no plan found, dbase still locked, fall through
        }
    }

    if (!insideP3(bpos)) {
        //cout<<"Goalie(.): position on P1."<<endl;
        goalieToInterceptPos(1);  // unlocks dbase
        return;
    } else { 
        //cout<<"Goalie(.): position on P0."<<endl;
        goalieToInterceptPos(0);  // unlocks dbase
        return;
    }
}


// 1.05o
// 3rd version of this, no enclosing while(1) loop, no intercept code, just
// dash-to's.
void GSPlayer::goalieCriticalActionLoop3() {
    
    //GDEBUG(cout<<"Goalie: entering critical loop."<<endl);
       
    // wait for the next tic, to give sensors change to catch up,
    // use no-mark version so we don't wait twice.
    waitNextFootCommandTicNoMark();  
        
    if (!goalieCriticalGameMode()) {
        //GDEBUG(cout<<"Goalie: Non-critical mode, exiting loop."<<endl);
       //V.1.06 : goalie always use narrow view
       // sendChangeView(SC_VW_normal, SC_VQ_high);
        waitNextFootCommandTic();  // do nothing
        return;
    }

    commandLock();
    ScPos bpos = ballPosAbs();
        
    if (!ballPosValid(5)) {
        GDEBUG(cout<<"Goalie(-): don't know where ball is!!"<<endl);
        commandUnlock();
	//V.1.06 : goalie always use narrow view
        //sendChangeView(SC_VW_normal, SC_VQ_high);
        turnToAbs(bpos);  // guess it was there ...
        for (int i = 0; i < 6; i++) {
            waitForVisualInfo();
            if (ballPosValid()) {
                cout<<"Goalie(-): found ball."<<endl;
                nodaNop() ;
                break;
            }
            turnTo(scNormalizeAngle(i*60 - 90));
        }
        GDEBUG(cout<<"Goalie(-): looked around for ball."<<endl);
        return;
    }
        
    if (bpos.x > 0) {
        //cout<<"Goalie: not on our side, exiting critical loop."<<endl;
        commandUnlock();
	//V.1.06 : goalie always use narrow view
        //sendChangeView(SC_VW_normal, SC_VQ_high);
        waitNextFootCommandTic();  // do nothing
        return;
    }

    // Case 1x: ball is close-in to our position
    if (goalieCriticalActionLoopBallIsClose()) {
        // action taken, dbase is unlocked by the above call
        return;
    } else {
        ; // no action taken, dbase is still locked, fall through
    }

    // Case 2, ball inside P2.  Run towards it.
    if (insideOurPenaltyBox(bpos)) {
        commandUnlock();
        stepDashToBall2(goalieInterceptDashPower, ScFalse);
        return;
    }


    if (!insideP3(bpos)) {
        //cout<<"Goalie(.): position on P1."<<endl;
        goalieToInterceptPos(1);  // unlocks dbase
        return;
    } else { 
        //cout<<"Goalie(.): position on P0."<<endl;
        goalieToInterceptPos(0);  // unlocks dbase
        return;
    }
}



// 1.04o
// for when the ball is within catchable range of us
// dbase is locked upon entry,
//     returns LOCKED if no action (returns ScFalse),
//             UNLOCKED if action taken (returns ScTrue).
ScBool GSPlayer::goalieCriticalActionLoopBallIsClose()
{
    ScPos bpos = ballPosAbs();
    if ( ((bpos - xposdir.pos).norm() >= SC_CATCHABLE_AREA_L)
         ||  ((bpos - eposdir.pos).norm() >= SC_CATCHABLE_AREA_L) ) {
        // ball isn't near enough for this code
        return ScFalse;
    }

    ScFloat dir;  // relative ball angle to body direction
    ScUInt tmargin = scycle.count - ball().ptime;
    ScPos bPosRel_e = bpos - eposdir.pos;
    ScPos bVelRel_e = ballVelAbs() - evel;
    ScPos ballPP1 = ballPredictedPosition(1);
    ScPos ballPP2 = ballPredictedPosition(1);

    //
    // If we have very very recent visual info and can catch,
    // do catch using that info
    //
    //if (tmargin <= 1  &&  !catchBanP()  ) {
    if (tmargin == 0  &&  !catchBanP()) {
        dir = bpos.AngleFrom(eposdir.pos, eposdir.bdir);
        
        //if (tmargin == 0) {
        //    // just saw ball, use epos
        //    dir = bpos.AngleFrom(eposdir.pos, eposdir.bdir);
        //} else if (tmargin == 1
        //           && getLastAction() != GSTurn
        //          && ballVelValid() 
        //           && bPosRel_e.norm() >= SC_CATCHABLE_AREA_W/2.0
        //          && bVelRel_e.norm() <= SC_CATCHABLE_AREA_W/2.0) {
        //    // just saw ball, and it should still be in the area
        //    dir = bpos.AngleFrom(eposdir.pos, eposdir.bdir);
        //}
        commandUnlock();
        GDEBUG(cout<<"Goalie(1a): grabbing ball from einfo."<<endl);
        catchBall(dir);
        return ScTrue;
    }

    //
    // Otherwise, we'll have to trust our xinfo.  If we can
    // catch, do so.
    //            
    if (CatchableP()) {
        dir = ballAngleRelBodyDir();
        commandUnlock();
        GDEBUG(cout<<"Goalie(1b): grabbing ball from xinfo."<<endl);
        catchBall(dir);
        return ScTrue;
    }

    //
    // So, we can't catch in this time step.  If we can kick,
    // do so.
    //
    if (KickableP()) {
        ScFloat kickADir = 60;
        if (bpos.x < 0) kickADir = -kickADir;
        commandUnlock();
        GDEBUG(cout<<"Goalie(1c): kicking out."<<endl);
        kickTo2(100, kickADir);
        return ScTrue;
    }

    //
    // The ball isn't kickable.  If we can catch in the next
    // timestep, just wait.  Maybe we'll get lucky and get a
    // vision update.
    //
    ScBool canCatch1 = (ballVelValid()
                        &&  ballPP1.x > -SC_PITCH_LENGTH/2.0
                        &&  ((ballPP1 - myPredictedPosition(1)).norm() <
                             SC_CATCHABLE_AREA_L) );
    if (catchBanP()  &&  !catchBanP(1)  &&  canCatch1) {
        commandUnlock();
        lastAction = GSNone;
        GDEBUG(cout<<"Goalie(1d): can't kick/catch now;"
               " try catch on next turn."<<endl);
        waitNextFootCommandTic();
        return ScTrue;
    }

    //
    // Ball isn't kickable, and we can't catch for awhile.
    // Dash to ball, in anticipation of kick.  
    //
    GDEBUG(cout<<"Goalie(1e): move towards ball."<<endl);
    if (ballVelValid()) {
        // turn/dash to where we expect the ball to be
        if (absAngleDiff(xposdir.pos.AngleTo(ballPP1), xposdir.bdir) < 10) {
            commandUnlock();
            dash(100);
            return ScTrue;
        } else {
            commandUnlock();
            turnTo(xposdir.pos.AngleTo(ballPP2));
            dash(100);
            return ScTrue;
        }
    } else {
        // just suppose ball is standing still
        if (absAngleDiff(ballAngleRel(), xposdir.bdir) < 10) {
            commandUnlock();
            dash(100);
            return ScTrue;
        } else {
            commandUnlock();
            turnTo(ballAngleRel());
            dash(100);
            return ScTrue;
        }
    }
    
    return ScFalse;  // we won't get here, but oh well
}


// 1.04o
ScBool GSPlayer::goalieCriticalGameMode() {
    if (playmode->id == SC_WORD(PlayOn).id  ||
        playmode->id == SC_WORD(DropBall).id) return ScTrue;

    switch (side) {
        case SC_S_left: {
            if (playmode->id == SC_WORD(KickInR).id
                || playmode->id == SC_WORD(CornerKickR).id
                || playmode->id == SC_WORD(FreeKickR).id) return ScTrue;
            break;
        }
        case SC_S_right: {
            if (playmode->id == SC_WORD(KickInL).id
                || playmode->id == SC_WORD(CornerKickL).id
                || playmode->id == SC_WORD(FreeKickL).id) return ScTrue;
            break;
        }
        default: {
            cerr << "Bad mojo."<<endl;
            exit(-1);
        }
    }
    return ScFalse;
}


//--------------------------------------------------
// step Swing


ScBool GSPlayer::stepSwing(ScBool righthandp) {
    if(!kickableP()) return ScFalse;

    ScPos rbpos = ball().posdir.pos.relativePosFrom(eposdir.pos);
    ScFloat dist = rbpos.norm();
    ScFloat adir = rbpos.Angle();
    ScFloat ddir = ((dist <= SC_BALL_SIZE + SC_PLAYER_SIZE) ?
                    90.0 : 
                    scASin((SC_BALL_SIZE + SC_PLAYER_SIZE)/dist));
    ddir += 15.0;

    ScFloat kadir =
        scNormalizeAngle(adir + 180 + (righthandp ? -ddir : ddir));

    kickTo(10,kadir);
    kickTo(-9,kadir);

    return ScTrue;
}


// deprecated
ScBool GSPlayer::stepBallToAbsAngle(ScFloat degAngle, ScBool &inSpotP)
{
    ScFloat ballAbsAngle;
    
    if (!KickableP()) {
        inSpotP = ScFalse;
        return ScFalse;
    }

    degAngle = scNormalizeAngle(degAngle);
    ballAbsAngle = ball().posdir.pos.AngleFrom(xposdir.pos);
    // 0 if ball is toward the Op-Goal side

    if (absAngleDiff(ballAbsAngle,degAngle) <
          GSCloseEnoughAngle) {
        inSpotP = true;
        return ScTrue;
    } else {
        inSpotP = false;
        // face ball
        turnToAbs(ball().posdir.pos);
        ScBool direction = scNormalizeAngle(degAngle - xposdir.bdir) > 0;

        cout << (direction ? "R " : "L ") << flush;
        
        return stepSwing(direction);
    }
}




//--------------------------------------------------
// dribbling

//
// ToDo: work with BallIsControllableP(), broadcasting..., high-level
// interrupt.
//
ScBool GSPlayer::simpleDribble(const ScPos& dest, ScFloat relAngle,
                               ScFloat kickPow, ScFloat dashPow,
                               ScInt stepCap)
{
    ScBool didSomethingP = false;
    ScBool inSpotP = false;
    ScUInt finalStep = scycle.count + stepCap;
    ScBool continueFlag;
    ScFloat absAngle;

    cout << "Dribble to " << dest << " : " << flush;
    
    while (scycle.count <= finalStep) {
        //
        // run to ball until it is kickable
        //
        while (!KickableP() && scycle.count <= finalStep) {

            cout << "D " << flush;
            
            // may need to get an actual lock on ball position here
            if (stepDashToBall(dashPow)) didSomethingP = true;
        }
        if (scycle.count > finalStep) break;

        //
        // move ball to angle
        //        
        inSpotP = continueFlag = false;
        absAngle = dest.AngleFrom(xposdir.pos) + relAngle;

        cout << endl << "(Absangle: " << absAngle << ") " << endl; 
       
        while (!inSpotP  &&  scycle.count <= finalStep) {
            if (stepBallToAbsAngle(absAngle, inSpotP)) {
                
                cout << "A " << flush;
                
                didSomethingP = true;
            } else {
                // server failed, or we're not close enough
                continueFlag = true;
                
                cout << "feh " << flush;
                
                break; // to _a_
            }            
        }
        // _a_
        if (scycle.count > finalStep) break;
        if (continueFlag) continue; // try to get close to ball again

        //
        // kick
        //
        while (scycle.count <= finalStep) {
            if (kickToAbs(kickPow, dest)) {
                cout << "K " << flush;
                didSomethingP = true;
                break;
            }
        }
        if (scycle.count > finalStep) break;
    } // loop to start of while

    cout << endl;

    return didSomethingP;
}


//--------------------------------------------------
// find clear way

ScFloat GSPlayer::findClearWay(const ScFloat& dir, const ScFloat& dir_step,
                               const ScFloat& dist, const ScFloat& width,
                               const ScInt& tmargin) {
    for(ScFloat ddir = 0.0; scAbs(ddir) <= 180.0; ddir += dir_step) {
        ScFloat d;
        d = scNormalizeAngle(dir+ddir);
        if(!isObstructed(d,dist,width,tmargin)) 
            return d;
        d = scNormalizeAngle(dir-ddir);
        if(!isObstructed((dir - ddir),dist,width,tmargin)) 
            return d;
    }
    return dir;
}


//--------------------------------------------------
// util functions




// eof //
