// -*- Mode: c++ -*-
#ifndef _GSPlayer_h_
#define _GSPlayer_h_
//
//  File: GSPlayer.h
//  ----------------
//  Along with GSPlayer.cc, defines:
//      class GSPlayer
// 
// 
#include <libsclient.h>
#include <assert.h>
#include "GSMessages.h"
#include "GSMovingObject.h"


//#define PLAYER_DEBUG
//#define GOALIE_DEBUG

#ifdef PLAYER_DEBUG
#define PDEBUG(x) x
#else
#define PDEBUG(x)
#endif

#ifdef GOALIE_DEBUG
#define GDEBUG(x) x
#else
#define GDEBUG(x)
#endif


// 1.04o
#define DB_LOCK(x) commandLock(); x; commandUnlock();

#define GS_SAY_SEPCHAR	'/'


typedef enum GSAction_T {
    GSNone,
    GSKick,
    GSTurn,
    GSDash,
    GSCatch,  // 1.04o
    GSError,
} GSAction;




typedef struct GSCommandStep_T {
    GSAction action;   // Which command (dash, kick, turn) to send.
    ScFloat direction; // Raw direction value to send.
    ScFloat power;     // Raw power value to send.
    
    //ScPos nextBallPos;  // in the case of a kick, use these values to 
    //ScPos nextBallVel;  // update our estimates, if the kick went through.
} GSCommandStep;


// circle kick directions
typedef enum GSRotationType_T {
    GSRotRight,             // clock-wise
    GSRotLeft,              // counter-clock-wise
    GSRotClosest,           // such that ball moves the least
    GSRotFurthest,          // such that ball opp GSRotClosest
    GSRotAvoid              // such that ball moves away from
    // nearest threatening opponent.
} GSRotationType;


// misc inline util functions defined at end of file
ScFloat absAngleDiff(ScFloat a, ScFloat b);
ScBool approxEqualAngle(ScFloat a, ScFloat b);
ScBool rightPCalculateHalfway(ScFloat fromAngle, ScFloat toAngle,
                              ScFloat *pNearAnswer, ScFloat *pFarAnswer);
inline ScBool posInBoundsP(const ScPos& p);


/* For keeping estimates up-to-date, several issues.
 *
 * Incoming visual data, as far as player's info (vel, pos, etc) is
 * concerned, only updates x-info if the sensed info is known to be
 * valid, otherwise estimates are kept.  Validity is determined by the
 * time at which the last command was _sent_.  If sent 100+ ms ago,
 * all visual info can be trusted.  Otherwise, what can be trusted
 * depends on what command was sent.
 * 
 * Incoming visual data always overwrites non-player info (ball vel,
 * pos, etc).  So when a kick command is sent, a timestamp is taken
 * of the balls esti-info (pos, vel), and this remains valid over any
 * visual data for 100 ms from the timestamp.
 */


class GSPlayer : public ScPlayer {
public:

    //--------------------------------------------------
    // misc parameters.  many of these values from CMUnited code

    // for turn-steps, successfull when within this value of desired dir.
    static const ScFloat GSCloseEnoughAngle = 10.0;
    
    // max number of elapsed ticks for pos/vel estimates to still be valid
    static const ScUInt GSBallConfTicksPos = 3;
    static const ScUInt GSBallConfTicksVel = 3;
    static const ScUInt GSPlayerConfTicks = 5;

    // how close for an opponenet to be considered a threat for
    // turning avoidance
    static const ScFloat GSAvoidDistance = 10.0;

    // how close is too close (from player center) for turnkicks
    static const ScFloat GSCloseMargin = (SC_PLAYER_SIZE +
                                           2.0 * SC_BALL_SIZE);
    
    // tweaked values to increase turnkick control
    static const ScFloat GSCircleKickFactor = 1.0;
    //static const ScFloat GSCircleKickFactor = .8;
    static const ScFloat GSCarefulCircleKickFactor = 1.0/3.0;
    static const ScFloat GSMaxCircleKickPower = 20.0;
    static const ScFloat GSCircleKickCloseEnoughAngle = 20.0;
    static const ScFloat GSOptControlDist = (SC_PLAYER_SIZE
                                              + .8 * SC_KICKABLE_MARGIN);
    
    // below this speed/velocity, the ball is considered "stopped"
    static const ScFloat GSBallVelDontCare = .05;
    
    // safer estimate of kickable area
    //static const ScFloat GSKickable = SC_KICKABLE_AREA - SC_BALL_SIZE;
    static const ScFloat GSKickable = SC_KICKABLE_AREA;
    
    // safer estimate of kickable area when dashing to ball
    static const ScFloat GSDashKickable = (SC_PLAYER_SIZE
                                            + .7 * SC_KICKABLE_MARGIN);

    // for smartKickTo, this determines whether we're "facing" the ball
    // or not.
    static const ScFloat GSSmartKickFaceBallAngle = 45;

    // for goalie, radius of perimeters (centered on GS_OurGoalCenter).
    static const ScPos GS_OurGoalCenter; // center of our goal
      //V.1.06: 
      // for more offensive defense
      //static const ScFloat GS_P0 = SC_GOAL_WIDTH / 4.0;
      //static const ScFloat GS_P1 = SC_GOAL_WIDTH / 2.0;
      //static const ScFloat GS_P2 = SC_PENALTY_AREA_LENGTH;
      static const ScFloat GS_P0 = SC_GOAL_WIDTH / 2.0;
      static const ScFloat GS_P1 = SC_PENALTY_AREA_LENGTH;
      static const ScFloat GS_P2 = SC_PENALTY_AREA_WIDTH / 2.0;
      static const ScFloat GS_P3;
      static const ScFloat GS_P4 = SC_PITCH_WIDTH / 2.0;  // 1.05o

    //--------------------------------------------------
    // members

    // moving-object database, and lock
    ScMutex objectTableMutex;
    GSMovingObjectTable objectTable;

    GSMessageRingBuffer messageRingBuffer;
    ScString sayBuffer;

    // we need a lock around estimateBodyState
    ScMutex estimateBodyStateMutex;
    
    // estimated ball pos/vel, not updated by sensor info (only by kick info)
    // and the timestamp/count when the estimate was made.
    ScPos estiBallPos;
    ScPos estiBallVel;
    ScCycleStep estiBallCycle;

    // cond var for syncing up command sends and body senses.
    ScMutexCondVar senseBodyFlag;
    ScMutexCondVar senseVisualFlag;
    ScMutexCondVar senseFullVisualFlag;

    // value of recovery factor [0,1]
    ScFloat recovery;

    // what action we last sucessfully sent (sent at fcycle.rtime);
    GSAction lastAction;

    // for waitNextFootCommandTic, bodySenseP is true iff we synced
    // from a sense-body signal the last time a command was sent.
    ScBool senseBodySignalP;

    // for waitNextFootCommandTic, if senseBodySignalP is false,
    // use this value instead.
    ScCycleStep waitCycle;
    
    // for position and covering area
    ScPos initPos ;
    ScPos homePos ;
    ScPos shiftPos ;
    ScPos areaLower ;	
    ScPos areaUpper ;

    ScInt closestCheckTimeMargin ;

    // 1.04o
    // goalie stuff
    
    // used for catchban predicates
    ScInt   goalieLastCatchTime;  

    // used for catch or kick intercept calculations
    ScInt   goalieMaxStepsForIntercept;

    // used when dashing to do a catch or kick intercept
    ScFloat goalieInterceptDashPower;

    // used when moving to position on P0/P1, and when moving towards
    // ball when not doing an actual intercept.
    ScFloat goalieMiscDashPower;
    
    // the goalie considers the ball "free" in P2 when it its moving
    // with velocity.norm() less than this value.
    ScFloat goalieFreeBallInP2MaxSpeed;

    ScFloat ballSeenRelDist ;
    ScFloat ballSeenRelAngToBDir ;
    ScInt ballSeenCycle ;

      // V.1.05noda
    ScCycleStep bdirLastUpdateByVisual ;

    //--------------------------------------------------
    // constructor

    GSPlayer() {
        sayBuffer.alloc(SC_BUFSIZE);
        recovery = 1.0;
        lastAction = GSNone;
        senseBodySignalP = ScTrue;
        estiBallCycle.update(); // mark initial estimate
        closestCheckTimeMargin = 5 ;

        // extra goalie stuff
        goalieLastCatchTime = 0;
        goalieMaxStepsForIntercept = 15;
        goalieInterceptDashPower = 100;
        goalieMiscDashPower = 100;
        goalieFreeBallInP2MaxSpeed = .5;
    };

    virtual ~GSPlayer();
    

    //--------------------------------------------------
    // init
    
    // halfTimeReset (gaea)
    void halfTimeReset() {
        recovery = 1.0;
        lastAction = GSNone;
    };

    
    //--------------------------------------------------
    // waits

    ScBool waitForVisualInfo() {
       // V.1.05noda
        return senseVisualFlag.timedlockwait(1);
    };

    // 1.04o this signal may never come if we're out of bounds,
    // since we can't get a fix on position.
    ScBool waitForFullVisualInfo() {
        //return senseFullVisualFlag.wait(); // 1.05o comment out
        return senseFullVisualFlag.timedlockwait(2); // 1.05o
    };


    //--------------------------------------------------
    // overridden basic server calls
    
    // Waits for signal from body-sense (returns true), or a time-out
    // period has elapsed (returns false)
    //
    // the "no mark" version just does the wait, nothing is changed.
    // So calling waitNextFootCommandTic() twice will wait twice,
    // calling the no-mark version and then waitNextFootCommandTic()
    // will only wait the first time.
    ScBool waitNextFootCommandTic();
    ScBool waitNextFootCommandTicNoMark();

    // Waits for next send-window, and sends command (all in gaea)
    ScBool turn(const ScFloat& pow);
    ScBool dash(const ScFloat& pow);
    ScBool kick(const ScFloat& pow, const ScFloat& dir);
    ScBool catchBall(const ScFloat& dir);
    ScBool turnTo(const ScFloat& dir);  // 1.05
    ScBool turnToRecursive(ScFloat adir, ScFloat speed);  // 1.05

    // these noLock variants are called by the above versions,
    // and do the actual calculations to update the x-info
    ScBool noLockTurn(const ScFloat& pow);
    ScBool noLockDash(const ScFloat& pow);
    void updateDashValues(const ScFloat& pow);
    ScBool noLockKick(const ScFloat& pow, const ScFloat& dir);
    void updateBallEstiInfo(ScFloat power, ScFloat rdir);

    GSAction getLastAction() {
        if (  vcycle.rtime.MSecFloat() - fcycle.rtime.MSecFloat()
              > simulationTic.MSecFloat()  ) {
            return GSNone;
        }
        return lastAction;
    }
    

    //--------------------------------------------------
    // object table and world-model

    GSMovingObject& ball() {
        return objectTable.getBall();
    };
    
    // gaea ?
    ScInt recvScanSInfoAll();
    // gaea
    ScBool recvScanSInfo();

    ScBool updateXValuesFromEValues();  // true iff full update


    //-------------
    // Functions that should be used prior to doing any caclculations
    // for 1-stepper actions.  Since the calcuations are fast, we can
    // wait for the next window, then lock the database, do
    // calculations, and immediatly send off the calculated command.
    //
    void commandWaitAndLock() {
        waitNextFootCommandTic();
        commandLock();
    };

    void commandLock() {
        estimateBodyStateMutex.lock();
        xposMutex.lock();
        objectTableMutex.lock();
    };

    void commandUnlock() {
        objectTableMutex.unlock();
        xposMutex.unlock();
        estimateBodyStateMutex.unlock();
    };


    //--------------------------------------------------
    // ball utils

    // Note, there is some prediction/estimation logic, to handle the
    // case when visual info may not reflect the kick that was just
    // (successfully) sent.  Info validity is given by BallPosValid()
    // and BallVelValid().
    //
    // The "right stuff" for prediction/estimation happens iff
    // BallPosAbs() and BallVelAbs() are used instead of
    // ball().posdir.pos and ball().vel.

    ScFloat ballPosUpdateTime() {  // float since .count is unsigned
        return scMax(ball().ptime, estiBallCycle.count);
    };

    ScFloat ballVelUpdateTime() {  // float since .count is unsigned
        return scMax(ball().vtime, estiBallCycle.count);
    };
    
    // 1.05o : made it so that parameter could be specified.
    ScBool ballPosValid(ScInt ticks = GSBallConfTicksPos) {
        // the .count value is unsigned...
        return (  (ScFloat)scycle.count - ballPosUpdateTime() <= ticks );
    };

    ScBool ballVelValid(ScInt ticks = GSBallConfTicksVel) {
        // the .count value is unsigned...
        PDEBUG(cout
               <<"\nBVV: scycle.count: "<<scycle.count
               <<"\nBVV: velupdate:    "<<ballVelUpdateTime()<<endl;);
        return ((ScFloat)scycle.count - ballVelUpdateTime() <= ticks );
    };

    // ballPosAbs (gaea)
    // -----------------
    // Frame of ref is the field.  Valid iff ballPosValid() says so.
    //
    ScPos ballPosAbs() {

        PDEBUG(
            cout
            << "\n    BPA: "
            << "\n         estiTime: "<<estiBallCycle.count
            << "\n         vTime:    "<<ball().ptime
            << "\n         estiPos:  "<<estiBallPos
            << "\n         vPos:     "<<ball().posdir.pos
            << endl;
            );
            
        
        ScPos ballPos;
        // use most recent ballPos.
        if (estiBallCycle.count > (ScUInt) ball().ptime) {
            PDEBUG(cout<<"BPA(e) " );
            ballPos = estiBallPos;
        } else {
            //PDEBUG(cout<<"BPA(n) ");
            ballPos = ball().posdir.pos;
        }

        // extrapolate, if possible, where it is now.
        if (ballVelValid()) {
            ScInt diff = (ScInt) (scycle.count - ballPosUpdateTime());
            
        PDEBUG(
            cout
            << "\n    BPA diff:   "<<diff
            << "\n        bpos:   "<<ballPos
            << "\n        bvel:   "<<ballVelAbs()
            << "\n        result: "
            <<predictedPosition(ScTrue, ballPos, ballVelAbs(), diff)
            <<endl;
            );
        
            return predictedPosition(ScTrue, ballPos, ballVelAbs(), diff);
        } else {
            return ballPos;
        }
    };

    // ballVelAbs (gaea)
    // -----------------
    // Frame of ref is the field.  Valid iff ballVelValid() says so.
    // Note: we always use estimate info over visual info for the
    // critical time periods (i.e. for 1.1*simulationTic.MSecFloat()
    // after we make an estimate).
    //
    ScPos ballVelAbs() {

        ScFloat diff = vcycle.rtime.MSecFloat() -
            estiBallCycle.rtime.MSecFloat();
        
        if ( diff < 1.1 * simulationTic.MSecFloat()  ||  // 10% buffer 
             estiBallCycle.count > (ScUInt) ball().vtime ) {
            PDEBUG(cout<<"BVA(e) ");
            return estiBallVel;
        } else {
            //PDEBUG(cout<<"BVA(n) ");
            return ball().vel;
        }
    };
    

    // position of ball, player at origin
    ScPos ballPosRel() {
        return ballPosAbs().relativePosFrom(xposdir.pos);
    };
    
    // vel of ball, player is ref-frame
    ScPos ballVelRel() {
        return ballVelAbs().relativePosFrom(xvel);
    };
    
    // direction of ball, player at origin, 0 deg is towards opp. goal
    ScFloat ballAngleRel() {
        return ballPosAbs().AngleFrom(xposdir.pos);
    };
    
    // direction of ball, player at origin, 0 deg body dir.
    ScFloat ballAngleRelBodyDir() {
        return ballPosAbs().AngleFrom(xposdir.pos, xposdir.bdir);
    };

    ScFloat distanceToBall() {
        return ballPosRel().norm();
    };
    
    ScBool kickableP() {
        //ScFloat eDistance = scDistance(eposdir.pos, ball().posdir.pos);

        //V.1.05 temp
//        return (distanceToBall() < GSKickable
//                &&  eDistance < GSKickable) ;
//                && kickablePbySensor()) ;
        return kickablePbySensor() ;
    };

    // V.1.05
    ScBool kickablePbySensor() {
        return (ballSeenRelDist < GSKickable
                &&  scycle.count - ballSeenCycle <= GSBallConfTicksPos);
    };
    // V.1.05
    ScBool dashKickablePbySensor() {
        return (ballSeenRelDist < GSDashKickable
                &&  scycle.count - ballSeenCycle <= GSBallConfTicksPos);
    };

    // gaea
    ScBool KickableP(ScUInt tmargin = GSBallConfTicksPos) {
        // .count is unsigned
        if((ScFloat)scycle.count - ballPosUpdateTime() > tmargin) {
            PDEBUG(
                cout
                <<"KickableP: false, not recent enough: "
                << (ScFloat)scycle.count - ballPosUpdateTime() <<endl;
                );
            return ScFalse;
        }
        if (kickableP()) {
            PDEBUG(cout<<"KickableP: true."<<endl);
            //PDEBUG(kickInfos("KickableP:", tmargin));
            return ScTrue;
        } else {
            PDEBUG(cout<<"KickableP: false: not close enough."<<endl;);
            PDEBUG(kickInfos("KickableP:", tmargin));
            return ScFalse;
        }
    };


    // gaea
    ScBool DashKickableP(ScUInt tmargin = GSBallConfTicksPos) {
        // .count is unsigned
        if((ScFloat)scycle.count - ballPosUpdateTime() > tmargin) {
            PDEBUG(
                cout
                <<"KickableP: false, not recent enough: "
                << (ScFloat)scycle.count - ballPosUpdateTime() <<endl;
                );
            return ScFalse;
        }
        //V.1.05 temp
//        if (distanceToBall() < GSDashKickable
//            &&  kickablePbySensor()) {
        if(dashKickablePbySensor()) {
            PDEBUG(cout<<"DashKickableP: true."<<endl;);
            PDEBUG(kickInfos("DashKickableP:", tmargin));
            return ScTrue;
        } else {
            PDEBUG(cout<<"DashKickableP: false, not close enough."<<endl);
            return ScFalse;
        }
    };

    
    ScFloat kickEffect() {
        if(!kickableP()) return 0.0;
        ScPos rbpos = ballPosRel();
        ScFloat margin = rbpos.norm() - SC_BALL_SIZE - SC_PLAYER_SIZE;
        ScFloat angle = scAbs(rbpos.Angle(xposdir.bdir));
        return (1.0 
                - 0.25 * (angle/180.0) 
                - 0.25 * (margin/SC_KICKABLE_MARGIN));
    };
    
    ScFloat KickEffect(ScUInt tmargin = GSBallConfTicksPos) {
        // .count is unsigned
        if((ScFloat)scycle.count - ballPosUpdateTime() > tmargin) return 0.0;
        else return kickEffect();
    };

    ScBool infrontKickableTo(const ScFloat& adir) {
        if(!kickableP()) return ScFalse;
        ScFloat rdir = ballPosAbs().AngleFrom(xposdir.pos,adir);
        return (scAbs(rdir) < 90.0);
    };

    ScBool infrontKickableToRel(const ScPos& rpos) {
        return infrontKickableTo(rpos.Angle());
    };

    ScBool infrontKickableToAbs(const ScPos& apos) {
        return infrontKickableToRel(apos.relativePosFrom(xposdir.pos));
    };

    //--------------------------------------------------
    // goalie functions

    // v1.04o
    // goalie function, TRUE iff goalie cannot catch ball because
    // of recent catch command.  Specify futuresteps = X for catch ban
    // status X steps into the future.
    ScBool catchBanP(ScInt futureSteps = 0) {
        if (playmode->id != SC_WORD(PlayOn).id) {
            PDEBUG(cout<<"catchBanP: FALSE, not normal play mode"<<endl);
            return ScFalse;
        }
        return ((ScFloat)scycle.count + futureSteps - goalieLastCatchTime
                < SC_CATCH_BAN_CYCLE);
    };        

    
    // v1.04o
    // goalie function
    ScBool catchableP() {
        if (distanceToBall() > SC_CATCHABLE_AREA_L) {
            PDEBUG(cout<<"catchableP: FALSE, not close enough"<<endl);
            return ScFalse;
        }
        if (catchBanP())  {
            PDEBUG(cout<<"catchableP: FALSE, catch-ban in effect"<<endl);
            return ScFalse;
        }
        return ScTrue;
    };

    
    // v1.04o
    // goalie function
    ScBool CatchableP(ScUInt tmargin = GSBallConfTicksPos) {
        if((ScFloat)scycle.count - ballPosUpdateTime() > tmargin)
            return ScFalse;
        return catchableP();
    };


    // v1.04o
    // primarily a goalie function, TRUE if the ball's velocity will
    // carry it into our goal.
    ScBool ballVelocityIntoMyGoalP() {
        if ( !ballPosValid() || !ballVelValid()) {
            PDEBUG(cout<<"ballVelIntoMyGoalP:  pos/vel not valid!"<<endl);
            return ScFalse;
        }
        
        ScPos bpos = ballPosAbs();
        ScPos bvel = ballVelAbs();
        // only worry about ball when it is in near-goal 1/3 of field,
        // and headed in our direction        
        if ( ballPosAbs().x > -SC_PITCH_LENGTH / 6.0  ||
             ballVelAbs().x > 0) {
            return ScFalse;
        }

        // solve for where it crosses our edge
        // point slope form: (edgeY-bpos.y)/(edgeX-bpos.x) = (bvel.y/bvel.x)
        ScFloat edgeY
            = (bvel.y/bvel.x)*(-(SC_PITCH_LENGTH/2.0)-bpos.x) + bpos.y;

        return (scAbs(edgeY) <= SC_GOAL_WIDTH/2.0 + 1.0); //+1 for safety
    };

    // 1.04o
    // goalie function
    ScFloat ballFreenessInGoalArea() {
        ScPos bpos = ballPosAbs();
        if ( !insideP3(bpos)  ||
             bpos.x > SC_PENALTY_AREA_LENGTH - SC_PITCH_LENGTH/2.0 )
            return 0.0;  // ball not close enough to goal

        if (playmode->id != SC_WORD(PlayOn).id)
            return 0.0;

        if (distanceToBall() < SC_CATCHABLE_AREA_L)
            return 1.0;  // up for grabs
        
        GSMovingObject *nearOpp
            = objectTable.nearestOpponentTo(bpos,
                                            scycle.count,
                                            closestCheckTimeMargin);
        if (scIsNullPtr(nearOpp)) {
            return 1.0; // no opponent, hmm
        }

        ScFloat oppDist = nearOpp->distanceTo(bpos) - SC_KICKABLE_AREA;
        ScFloat myDist = distanceToBall() - SC_CATCHABLE_AREA_L;

        if (myDist >= oppDist) {
            return 1.0;
        }

        return (oppDist / myDist);        
    };

    // 1.04o
    // goalie function, gives relative "freeness" of ball [0,1] given
    // how close nearest opponent is, and how close player is.  .5 means
    // the opponent is twice as close, .33 means three times as close,
    // taking into account the KICKABLE and CATCHABLE areas.
    ScFloat BallFreenessInGoalArea(ScUInt tmargin = GSBallConfTicksPos) {
        if((ScFloat)scycle.count - ballPosUpdateTime() > tmargin) {
            PDEBUG(
                cout
                <<"BallFreenessInGoalArea: false, not recent enough: "
                << (ScFloat)scycle.count - ballPosUpdateTime() <<endl;
                );
            return 0.0;
        }
        return ballFreenessInGoalArea();
    };

    
    //--------------------------------------------------
    // simple prediction

    ScPos myPredictedPosition(ScInt steps = 1) {
        return predictedPosition(ScFalse, xposdir.pos, xvel, steps);
    };

    ScPos myPredictedPositionWithDash(ScFloat power, ScInt steps = 1) {
        ScFloat depByPow = dashEffectivePower(power) * SC_DASH_POWER_RATE;
        ScPos movement = xvel + Polar2ScPos(depByPow, xposdir.bdir);
        ScPos newVel = movement * SC_PLAYER_DECAY;
        return predictedPosition(ScFalse, xposdir.pos+movement,
                                 newVel, steps-1);
    };

    ScPos ballPredictedPosition(ScInt steps = 1) {
        return predictedPosition(ScTrue, ballPosAbs(), ballVelAbs(), steps);
    };

    ScPos predictedPosition(ScBool ballP, ScPos pos,
                            ScPos velocity, ScInt steps)
    {
        ScFloat decay = (ballP) ? SC_BALL_DECAY : SC_PLAYER_DECAY;
        for (ScInt i = 0; i < steps; i++) {
            pos += velocity;
            velocity *= decay;            
        }
        return pos;
    };


    //--------------------------------------------------
    // middle command
        

    // executeGSCommandStep
    // --------------------
    // Sends calculated command to server.
    //
    ScBool executeGSCommandStep(const GSCommandStep& cs)
    {
        PDEBUG(cout << "\n\nexecuteGSCommandStep. " << endl);
        
        ScBool result = ScTrue;
        switch (cs.action) {
            case GSNone: {
                break;
            }
            case GSError: {
                result = ScFalse;
                break;
            }
            case GSKick: {
                result = noLockKick(cs.power, cs.direction);
                if (result) {
                    //estiBallPos = cs.nextBallPos;
                    //estiBallVel = cs.nextBallVel;
                    //estiBallCycle.update();
                    //estiBallCycle.count = scycle.count + 1;
                    // +1 because it reflects the next timestep,
                    // this is most important for BallPosAbs()
                    // and BallVelAbs().

                    //PDEBUG(
                    //    cout<<"executeGSCommandStep: "
                    //    "estiBallCycle update."<<endl;
                    //    );
                }
                break;
            }
            case GSTurn: {
                result = noLockTurn(cs.direction);
                break;
            }
            case GSDash: {
                result = noLockDash(cs.power);
                break;
            }
            default: {
                cerr << "Bad Mojo" << endl;
                exit(-1);
            }
        }

        PDEBUG(cout<<endl);
        
        return result;
    };


    //--------------------------------------------------
    // middle command: kicks
    
    // distanceKick (gaea)
    // -------------------
    // Kick such that ball moves in absolute direction 'absDir' for 'dist'
    // * 'dFactor' units (i.e., will be that far away from its current
    // position on the next tic).
    // 
    // Ball's current velocity is corrected for, if known.  If the
    // kick requirers too much power, kick is made at maxPower.  TRUE
    // is returned if action went through normally, FALSE if server
    // missed the command.  Note, no check is made to ensure that the
    // ball is not kicked into the player.
    //
    // NOTE: This is the only kick function which currently updates
    // the esti-ball estimates.
    //
    ScBool distanceKick(ScFloat absKickDir,
                        ScFloat kickDist,
                        ScFloat dFactor = 1.0,
                        ScFloat maxPower = SC_MAXPOWER)
    {
        GSCommandStep cs;
        ScBool result;
        commandWaitAndLock();
        calcDistanceKick(&cs, absKickDir, kickDist, dFactor, maxPower);
        result = executeGSCommandStep(cs);
        commandUnlock();
        return result;
    }

    // fills in the passed-in GSCommandStep structure
    void calcDistanceKick(GSCommandStep *pwo_cs,
                          ScFloat absKickDir, ScFloat kickDist,
                          ScFloat dFactor, ScFloat maxPower = SC_MAXPOWER);
    

    // circleKick (gaea)
    // -----------------
    // Uses distanceKick to kick the ball around the player,
    // maneuvering it towards the given absAngle (relative to player's
    // position, 0 is towards opponent goal).  The ball is at the
    // given absAngle when both this function and *pwo_DoneP are true.
    // Specify rightP==TRUE if the ball is to go in a clock-wise turn.
    //
    // Note: currently the stopBallP and getting the ball close to
    // absAngle are too noisy to be useful, however, the basic utility
    // of getting the ball from one side to the other (say, for a
    // pass) seems to work.
    //
    ScBool circleKick(ScFloat absAngle,
                      ScBool rightP,
                      ScBool stopBallP,
                      ScBool *pwo_DoneP)
    {
        return circleKick(absAngle, rightP, stopBallP, GSOptControlDist,
                          GSCloseMargin, GSCircleKickFactor, pwo_DoneP);
    };

    // Fully specified version.
    ScBool circleKick(ScFloat absAngle,
                      ScBool rightP,
                      ScBool stopBallP,
                      ScFloat endDist,
                      ScFloat closeMargin,
                      ScFloat kickFactor,
                      ScBool *pwo_DoneP,
                      ScBool controlCircleP = ScTrue)
    {
        GSCommandStep cs;
        ScBool result;
        commandWaitAndLock();
        calcCircleKick(&cs, absAngle, rightP, stopBallP, endDist,
                       closeMargin, kickFactor, pwo_DoneP, controlCircleP);
        result = executeGSCommandStep(cs);
        commandUnlock();
        return result;
    };

    // Fills in the passed-in GSCommandStep structure.
    void calcCircleKick(GSCommandStep *pwo_cs,
                        ScFloat absAngle,
                        ScBool rightP,
                        ScBool stopBallP,
                        ScBool *pwo_DoneP,
                        ScBool controlCircleP = ScTrue)
    {
         calcCircleKick(pwo_cs, absAngle, rightP, stopBallP, GSOptControlDist,
                        GSCloseMargin, GSCircleKickFactor, pwo_DoneP,
                        controlCircleP);
    };

    // Fully specified fill-in version.
    void calcCircleKick(GSCommandStep *pwo_cs,
                        ScFloat absAngle,
                        ScBool rightP,
                        ScBool stopBallP,
                        ScFloat endDist,
                        ScFloat closeMargin,
                        ScFloat kickFactor,
                        ScBool *pwo_DoneP,
                        ScBool controlCircleP = ScTrue);


    // multiPassBallTo (gaea)
    // ----------------------
    // Wrapper for passBallTo, one call does multiple actions.
    //
    // Kicks the ball at 'kickPow' power in 'adir' direction.  If the
    // ball is not in the correct hemisphere, circleKicks are done to
    // move the ball, such that the ball moves through the least
    // amount of angle.  Returns true if the pass happened, false if
    // it did not (eg, ball wasn't kickable, or lost ball).
    //
    ScBool multiPassBallTo(ScFloat kickPow, ScFloat adir,
                           ScBool faceBallP = ScFalse)
    {
        ScBool doneP;
        ScBool completionFlag = ScFalse;
        while (KickableP()) {
            PDEBUG(cout << "\nmultiPassBallTo: (kickable)\n"<<endl);
            if (passBallTo(kickPow, adir, &doneP, faceBallP)  &&  doneP) {
                PDEBUG(cout << "\npassBallTo: kick made.\n"<<endl);
                completionFlag = ScTrue;
                break;
            }
        }
        return completionFlag;
    };


    // passBallTo (gaea)
    // -----------------
    // Single-step action of multiPassBallTo; either kicks or circleKicks.
    //
    ScBool passBallTo(ScFloat kickPow, ScFloat adir,
                      ScBool *pwo_DoneP, ScBool faceBallP = ScFalse)
    {
        GSCommandStep cs;
        ScBool result;
        commandWaitAndLock();
        ScBool controlCircleP = ScTrue;  // slow version
        //ScBool controlCircleP = ScFalse; // fast version
        calcSmartKickTo(&cs, kickPow, adir, controlCircleP,
                        faceBallP, GSRotClosest, pwo_DoneP);
        result = executeGSCommandStep(cs);
        commandUnlock();
        return result;
    };


    ScBool multiSmartKickTo(ScFloat kickPow,
                            ScFloat adir,
                            ScBool controlCircleP,
                            ScBool faceBallP,
                            GSRotationType rotType)
    {
        ScBool doneP;
        ScBool completionFlag = ScFalse;
        while (KickableP()) {
            PDEBUG(cout << "\nmultiSmartKickTo: (kickable)\n"<<endl);
            if (  smartKickTo(kickPow, adir, controlCircleP,
                            faceBallP, rotType, &doneP)  &&
                  doneP  ) {
                PDEBUG(cout << "\nmultiSmartKickTo: kick made.\n"<<endl);
                completionFlag = ScTrue;
                break;
            }
        }
        return completionFlag;
    };

    
    // smartKickTo (gaea, but not yet)
    // ------------------
    // This code can be regarded as a more intelligent kickTo: Tries
    // to kick ball in 'adir' at 'kickPow'.  If the ball is not in the
    // correct hemisphere (kick could hit the player), then a
    // circleKick is used to move the ball around the player.
    //
    // 'controlCircleP' affects circleKick behavior.  If TRUE, then the
    // circleKicks are done more slowly, otherwise, the ball may be
    // kicked much (MUCH) harder than expected.
    //
    // 'faceBallP' specifies that the player should, if possible,
    // turn to the ball before making the final kick.  This seems
    // to work only when controlCircleP==TRUE, and even then,
    // not all that well.
    //
    // 'rotType' defines the direction in which circleKicks, if
    // needed, are done.
    //
    // The kick was presumably made when this function returns true
    // and *pwo_Done is true.
    //
    ScBool smartKickTo(ScFloat kickPow,
                       ScFloat adir,
                       ScBool controlCircleP,
                       ScBool faceBallP,
                       GSRotationType rotType,
                       ScBool *pwo_DoneP)
    {
        GSCommandStep cs;
        ScBool result;
        commandWaitAndLock();
        calcSmartKickTo(&cs, kickPow, adir, controlCircleP,
                        faceBallP, rotType, pwo_DoneP);
        result = executeGSCommandStep(cs);
        commandUnlock();
        return result;
    };

    // Fills in the passed-in GSCommandStep structure.
    void calcSmartKickTo(GSCommandStep *pwo_cs,
                         ScFloat kickPow,
                         ScFloat adir,
                         ScBool controlCircleP,
                         ScBool faceBallP,
                         GSRotationType rotType,
                         ScBool *pwo_DoneP);
        

    // kickTo2 (gaea, but not yet)
    // --------------
    // Similar to kickTo, kicks ball at 'kickPow' power in 'adir'
    // direction.  However, the ball's velocity, if known, is
    // corrected for, such that the ball's trajectory is as close to '
    // adir' as possible (if there is not enough power, it may not be
    // possible to kick in 'adir').  Regardless, the kick is made at
    // the given power.
    //
    // Note: no check is done to see if the ball will hit the player.
    //
    ScBool kickTo2(ScFloat kickPow, ScFloat adir)
    {
        GSCommandStep cs;
        ScBool result;
        commandWaitAndLock();
        calcKickTo2(&cs, kickPow, adir);
        result = executeGSCommandStep(cs);
        commandUnlock();
        return result;
    };

    // Fills in the passed-in GSCommandStep structure.
    void calcKickTo2(GSCommandStep *pwo_cs, ScFloat kickPow, ScFloat adir);
    
    
    //--------------------------------------------------
    // middle command: movement

    // stepGoto2 (gaea)
    // ----------------
    // 2-stepper: will either turn towards 'dest', or (if already
    // oriented in the right direction) will dash towards 'dest'.
    // Specify avoidp=TRUE to avoid running into things.
    //
    ScBool stepGoto2(const ScPos& dest, ScFloat dashPow, 
                     ScBool avoidp = ScTrue)
    {
        GSCommandStep cs;
        ScBool result;
        commandWaitAndLock();
        // begin 1.05o addition
        ScFloat distToDest = scDistance(xposdir.pos, dest);
        if (distToDest < 1.0) {
            // don't overshoot destination
            ScFloat dashPowToClose
                = ((distToDest-estimateBodyState.speed)
                   / SC_DASH_POWER_RATE * estimateBodyState.effort);
            dashPow = scMin(dashPow, dashPowToClose);
        }
        // end 1.05o addition
        calcStepGoto2(&cs, dest, dashPow, avoidp);
        result = executeGSCommandStep(cs);
        commandUnlock();
        return result;
    };
    
    // stepDashToBall2 (gaea)
    // ----------------------
    // Just like stepGoto2, except the destination is the ball
    // position.
    //
    // Note: *no* check is made to verify that the ball position is
    // valid.
    //
    ScBool stepDashToBall2(ScFloat dashPow, ScBool avoidp = ScTrue) {
        GSCommandStep cs;
        ScBool result;
        commandWaitAndLock();

        // begin 1.05 addition
        ScPos bpos = ballPosAbs();
        ScFloat distToBall = scDistance(xposdir.pos, bpos);
        if (distToBall < 1.0) {
            // don't overshoot ball
            ScFloat dashPowToClose
                = ((distToBall-estimateBodyState.speed)
                   / SC_DASH_POWER_RATE * estimateBodyState.effort);
            dashPow = scMin(dashPow, dashPowToClose);
        }
        // end 1.05 addition
        
        calcStepGoto2(&cs, ballPosAbs(), dashPow, avoidp);
        result = executeGSCommandStep(cs);
        commandUnlock();
        return result;
    };
    
    // fills in the passed-in GSCommandStep structure
    void calcStepGoto2(GSCommandStep *pwo_cs,
                       const ScPos& dest, ScFloat power, ScBool avoidp);
    

    //--------------------------------------------------
    // middle command: planning

    
    // getPlanForIntercept (gaea)
    // --------------------------
    // This is slightly different from the normal middle-commands, as
    // it doesn't execute anything.  Instead, it merely does calculations.
    //
    // BallPosValid() and BallVelValid() must both be true.
    //
    // Given the ball's current velocity under a zero-force
    // assumption, and the specified dashPow, this function will
    // return a turning direction, and the number of dashes (at that
    // power) to be executed after the turn to put the ball within
    // DashKickable distance (i.e., interceptable distance).
    //
    // ScTrue is returned if a valid intercept plan within maxSteps
    // is found.  If *pwo_numDashes == 0, then the ball is either
    // kickable, or will be in the next time-step (the turn direction
    // can be ignored, in this case).
    //
    // Note: This takes into account what the dash actually will
    // accomplish, given stamina, etc., but assumes inital player
    // velocity is zero.
    //
    // If extraSafety is true, then an intercept plan that arrives
    // ahead of the ball is attempted.
    //
    // NOTE: no LOCKING/UNLOCKING of database done, be sure to
    //       call this only when dbase is LOCKED.
    //
    // ---
    // Later implementation note: check for obstruction.
    // 
    // Later implementation note: in returning absTurnDir, we will
    // probably want to fix TurnTo to take into account our current
    // speed, so that the dashes are done in the right direction.
    //
    // Later implementation note: player will probably want to keep a
    // fix on the ball via head turning, in which case this function
    // can be called several times to tune the approach.
    //
    // v1.04o: made into wrapper for getPlanForIntercept2
    ScBool getPlanForIntercept(ScFloat dashPow,
                               ScInt maxSteps,
                               ScBool extraSafetyP,
                               ScFloat* pwo_absTurnDir,
                               ScInt* pwo_numDashes) {
        return getPlanForIntercept2(dashPow, maxSteps, extraSafetyP,
                                    0, 1, pwo_absTurnDir, pwo_numDashes,
                                    NULL, NULL, NULL);
    }

    //  v1.04o
    ScBool getPlanForIntercept2(ScFloat dashPow,
                                ScInt maxSteps,
                                ScBool extraSafetyP,
                                ScFloat margin,
                                ScFloat ballVelFactor,
                                ScFloat *pwo_absTurnDir,
                                ScInt *pwo_numDashes,
                                ScPos *pwo_endPlayerPos,
                                ScPos *pwo_endBallPos,
                                ScPos *pwo_endBallVel);

    // v1.04o
    // goalie function, dashes at goalieInterceptDashPower
    ScBool getPlanForGoalieCatchIntercept(ScFloat *pwo_absTurnDir,
                                          ScInt *pwo_numDashes,
                                          ScFloat *pwo_absCatchDir);

    // goalie stuff
    ScBool attemptGoalieCatchIntercept(ScBool makeCatchP);
    ScBool attemptGoalieKickIntercept(ScBool makeKickP);
    void goalieToInterceptPos(ScInt perimeter);
    void goalieCriticalActionLoop();
    void goalieCriticalActionLoop2(); // 1.05o
    void goalieCriticalActionLoop3(); // 1.05o
    ScBool goalieCriticalActionLoopBallIsClose();
    ScBool goalieCriticalGameMode();


    
    //--------------------------------------------------
    // middle command:

    ScBool stepSwing(ScBool righthandp = ScTrue);

    ScBool stepGoto(const ScPos& dest, ScFloat power, 
                    ScBool avoidp = ScTrue);
    
    ScBool stepDashToBall(ScFloat dashPow) {
        return stepGoto(ballPosAbs(), dashPow);
    };

    // If the ball is not close enough to the given abs angle (i.e.,
    // facing opponent side, 0 is forward, 90 is right, 180 is behind,
    // 270 is left), player turns to ball and swing-kicks.  False is
    // returned if ball isn't kickable, or if stepSwing fails.
    //
    // deprecated
    ScBool stepBallToAbsAngle(ScFloat degAngle, ScBool &inSpotP); 


    // This is a mid-level command, and operates for about "stepCap" steps.  
    //
    //     (1)  dash within range of the  ball.
    //     (2)  maneuver ball (approx) to given relative angle.
    //     (3)  kick ball with specified power in direction of dest.
    //     (4)  dash with specified power in direction parallel to the ball.
    //
    // and repeats until the number of steps has expired.  The
    // relative angle is the orientation of ball (0 is front, +/- are
    // R/L) along the player's intended path of travel.

    // deprecated
    ScBool simpleDribble(const ScPos& dest, ScFloat relAngle, ScFloat
                          kickPow, ScFloat dashPow, ScInt stepCap=5);

        
 
    //--------------------------------------------------
    // obstruction

    ScBool isObstructed(const ScFloat& dir,
                        const ScFloat& dist, const ScFloat& width,
                        const ScInt& tmargin = 1) {
        return objectTable.isObstructed(dir,dist,width,
                                        xposdir.pos,scycle.count,tmargin);
    };

    ScBool isObstructed(const ScPos& dest,
                        const ScFloat& dist, const ScFloat& width,
                        const ScInt& tmargin = 1) {
        return objectTable.isObstructed(dest,dist,width,
                                        xposdir.pos,scycle.count,tmargin);
    };

    //--------------------------------------------------
    // find clear way

    ScFloat findClearWay(const ScFloat& dir, const ScFloat& dir_step,
                         const ScFloat& dist, const ScFloat& width,
                         const ScInt& tmargin = 1);

    ScFloat findClearWay(const ScPos& dest, const ScFloat& dir_step,
                         const ScFloat& dist, const ScFloat& width,
                         const ScInt& tmargin = 1) {
        return findClearWay(xposdir.pos.AngleTo(dest),dir_step,
                            dist,width,tmargin);
    };

    //--------------------------------------------------
    // ball posession

    
    ScBool closestTeamMemberToBallP() {
        // PDEBUG(cout << "closestTeamMemberToBallP: \n");
        
        GSMovingObject* p
            = objectTable.nearestTeammateTo(ballPosAbs(), scycle.count,
                                            closestCheckTimeMargin);
        
        if(scIsNullPtr(p)) {
            // PDEBUG(cout<< "     no teammates!, I am closest." << endl);
            return ScTrue;
        }

        //PDEBUG(
        //    cout
        //    << "     teammate distance from ball: "
        //    << p->distanceTo(BallPosAbs()) << endl
        //    << "     my distance to ball: "
        //   << distanceToBall() << endl;
        //    );

        if (p->distanceTo(ballPosAbs()) > distanceToBall())  {
            // PDEBUG(cout<<"     --I am closest"<<endl);
            return ScTrue;
        }

        // PDEBUG(cout<<"     --I am NOT closest"<<endl);
        return ScFalse;
    };
    
    //--------------------------------------------------
    // misc helpers

    //-----------
    // "calculate" versions of basic commands
    
    // do nothing
    void calcNoAction(GSCommandStep *pwo_cs) {
        pwo_cs->action = GSNone;
    };

    // do nothing
    void calcError(GSCommandStep *pwo_cs) {
        pwo_cs->action = GSError;
    };
    
    // to absolute direction
    void calcTurnTo(GSCommandStep *pwo_cs, const ScFloat& dir) {
        pwo_cs->action = GSTurn;
        pwo_cs->direction = scNormalizeAngle(dir - xposdir.bdir);
    };

    // towards relative position
    void calcTurnToRel(GSCommandStep *pwo_cs, const ScPos& rpos) {
        calcTurnTo(pwo_cs, rpos.Angle()) ;
    };

    // towards absolute position
    void calcTurnToAbs(GSCommandStep *pwo_cs, const ScPos& apos) {
        calcTurnToRel(pwo_cs, xposdir.pos.relativePosTo(apos)) ;
    };

    // to abosolute direction
    void calcKickTo(GSCommandStep *pwo_cs,
                    const ScFloat& power, const ScFloat& dir) {
        pwo_cs->action = GSKick;
        pwo_cs->direction = scNormalizeAngle(dir - xposdir.bdir);
        pwo_cs->power = power;
    };

    // to relative position
    void calcKickToRel(GSCommandStep *pwo_cs, 
                       const ScFloat& power, const ScPos& rpos) {
        calcKickTo(pwo_cs, power, rpos.Angle()) ;
    };

    // to aboslute position
    void calcKickToAbs(GSCommandStep *pwo_cs,
                       const ScFloat& power, const ScPos& apos) {
        calcKickToRel(pwo_cs, power, xposdir.pos.relativePosTo(apos)) ;
    };


    //-----------
    // middle command util functions

    void calcStopBallKick(GSCommandStep *pwo_cs)
    {
        if (!KickableP()) {
            PDEBUG(cout<<"calcStopBallKick: not kickable?!?"<<endl);
            calcError(pwo_cs);
            return;
        }

        ScFloat kdir, kdist;
        // kick into self (too dangerous....)
        //kdir = ballPosAbs().AngleTo(xposdir.pos);
        //kdist = distanceToBall();
        kdir = kdist = 0;
        PDEBUG(cout<<"calcStopBallKick: pre-correction: dir: "<<kdir
               <<" dist: "<<kdist<<endl);
        playerMovementCorrectDirDist(&kdir, &kdist);
        PDEBUG(cout<<"calcStopBallKick: post-correction: dir: "<<kdir
               <<" dist: "<<kdist<<endl);
        calcDistanceKick(pwo_cs, kdir, kdist, 1.0);
    }
    
    void playerMovementCorrectDirDist(ScFloat *pwo_dir, ScFloat *pwo_dist);

    ScBool straightKickP(ScFloat degAngle, ScFloat dist,
                         ScFloat closeMargin);

    ScBool getCorrectTurnDir(GSRotationType rotType, ScFloat targetADir);


    //-----------
    // calculations for action updates

    ScFloat actualTurnAngle(const ScFloat& angle) {
        return angle / (1.0 + SC_INERTIA_MOMENT * estimateBodyState.speed);
    };

    ScFloat staminaConsumedByDash(const ScFloat& pow) {
        return (pow >= 0) ? (pow) : (-2.0 * pow);
    };

    // 1.04o change
    ScFloat dashEffectivePower(const ScFloat& pow) {
        ScFloat stamina = scMax((estimateBodyState.stamina -
                                 staminaConsumedByDash(pow)), 0);
        if (pow > 0) {
            return estimateBodyState.effort * scMin(pow, stamina);
        } else {
            return -(estimateBodyState.effort * scMin(-pow, stamina));
        }
    };

    void recoverEstimateStamina() {
        ScFloat oldStamina = estimateBodyState.stamina;
        estimateBodyState.stamina =
            scMin(SC_STAMINA_MAX, oldStamina + recovery * SC_STAMINA_INC_MAX);
    };


    // 1.04o , 1.05o
    ScBool insideP0(const ScPos& p) {
        return scDistance(GS_OurGoalCenter,p) < GS_P0;
    }
    ScBool insideP1(const ScPos& p) {
        return scDistance(GS_OurGoalCenter,p) < GS_P1;
    }
      // V.1.06 : Is it better to use insideOurPenaltyBox rather than insideP2??
    ScBool insideP2(const ScPos& p) {
//        return scDistance(GS_OurGoalCenter,p) < GS_P2;
       return scDistance(GS_OurGoalCenter,p) < GS_P2 || insideOurPenaltyBox(p) ;
    }
    ScBool insideP3(const ScPos& p) {
        return scDistance(GS_OurGoalCenter,p) < GS_P3;
    }
    ScBool insideP4(const ScPos& p) {
        return scDistance(GS_OurGoalCenter,p) < GS_P4;
    }
    ScBool insideCenterCorridor(const ScPos& p) {
        return (scAbs(p.x) < GS_P2);
    }
    ScFloat getPerimeterDistance(ScInt perimeterNumber) {
        switch (perimeterNumber) {
            case 0: return GS_P0;
            case 1: return GS_P1;
            case 2: return GS_P2;
            case 3: return GS_P3;
            case 4: return GS_P4;
            default: cerr<<"Bad mojo.\n"; exit(-1);
        }
        return 0; // for looks
    }
    ScBool insideOurPenaltyBox(const ScPos& p) {
        return ((scAbs(p.y) < SC_PENALTY_AREA_WIDTH / 2.0)
                && (p.x > 0 &&  p.x < (-SC_PITCH_LENGTH/2.0
                                       + SC_PENALTY_AREA_LENGTH)));
    }
                

    
    //--------------------------------------------------
    // debug
    
    void printPositionInfo(char *pre) {
        eposMutex.lock();
        xposMutex.lock();
        cout<< endl
            << pre << "bpos  : " << ballPosAbs() << endl
            << pre << "pos   : " << xposdir.pos << endl
            << pre << "dist  : " << distanceToBall() << endl
            << pre << "v-bpos: " << ball().posdir.pos << endl
            << pre << "v-pos : " << eposdir.pos << endl
            << pre << "v-dist: " << (eposdir.pos - ball().posdir.pos).norm()
            ;
        eposMutex.unlock();
        xposMutex.unlock();
    };

    void kickInfos(char *pre, ScUInt tmargin) {
        cout
            << pre << " infos:"<<endl
            << pre << "    scycle.count:      "<<scycle.count<<endl
            << pre << "    ballPosUpdateTime: "<<ballPosUpdateTime()<<endl
            << pre << "    tmargin:           "<<tmargin<<endl
            << pre << "    MyPos:             "<<xposdir.pos<<endl
            << pre << "    BallPos:           "<<ballPosAbs()<<endl
            << pre << "    distToBall:        "<<distanceToBall()<<endl
            << pre << "    ballVel:           "<<ballVelAbs()<<endl
            ;
    };

    
    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#GSPlayer[" << teamname
          << ":" << unum << "/" << scSide2Char(side) << "]";
        if(detailp) {
            ScPlayer::describe(o,detailp,newlinep);
        } else {
            if(newlinep) o << endl;
        }
    };

    friend ostream& operator<< (ostream& o, const GSPlayer& p) {
        p.describe(o,ScFalse,ScFalse);
        return o;
    };
    
};  // end of class GSPlayer



//----------------------------------------------------------------------
// misc inline util functions

inline ScFloat absAngleDiff(ScFloat a, ScFloat b)
{
    return scAbs(scNormalizeAngle(a-b));
}

inline ScBool approxEqualAngle(ScFloat a, ScFloat b)
{
    return (absAngleDiff(a,b) < 1.0);
}

// If the ball is at fromAngle, and we want it at toAngle, there
// are two halfway points: a near one and a far one.
//   Let dNear = scAbs(scNormalizeangle(fromAngle-NearAnswer)), and
//   let dFar  = scAbs(scNormalizeangle(fromAngle-FarAnswer)), then
//      (2 * dNear) <= 180,  and
//      (2 * dFar)  >= 180
//
// ScTrue is returned if the nearAnswer is in a rightward turning
// direction.  Note that the FarAnswer and nearAnswer are just 180 deg
// opposite.  The values for the the near and far answers are written
// in if the pointers are non-NULL.
inline ScBool rightPCalculateHalfway(ScFloat fromAngle, ScFloat toAngle,
                                     ScFloat *pNearAnswer,
                                     ScFloat *pFarAnswer)
{
    ScFloat nearAnswer;
    ScBool rightP;

    fromAngle = scNormalizeAngle(fromAngle);
    toAngle = scNormalizeAngle(toAngle);

    ScFloat angleDiff = scAbs(scNormalizeAngle(fromAngle - toAngle));
    // angleDiff is from [0,180]
    
    if (approxEqualAngle((fromAngle + angleDiff), toAngle)) {
        nearAnswer = scNormalizeAngle(fromAngle + angleDiff / 2.0);
        rightP = ScTrue;
    } else {
        nearAnswer = scNormalizeAngle(fromAngle - angleDiff / 2.0);
        rightP = ScFalse;
    }
    if (pNearAnswer != NULL) *pNearAnswer = nearAnswer;
    if (pFarAnswer != NULL) *pFarAnswer = scNormalizeAngle(nearAnswer + 180);
    return rightP;
}

// 1.04o
inline ScBool posInBoundsP(const ScPos& p) {
    return (scAbs(p.x) < SC_PITCH_LENGTH/2.0  &&
            scAbs(p.y) < SC_PITCH_WIDTH/2.0);
}
// eof //
#endif
